"""
Moving Average trading signals

"""

import numpy as np
from technicalmethods.methods import Indicators

class MovingAverageEntry():
    """
    Functions to create moving average crossover trading signals

    """
    @staticmethod
    def entry_double_ma_crossover(prices, params):
        """
        Entry signal for Double Moving Average Crossover strategy

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        ma1 : Int
            The faster moving average.
        ma2 : Int
            The slower moving average.
        simple_ma : Bool
            Whether to calculate a simple or exponential moving average. The
            default is True.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals.

        """
        ma1 = params['ma1']
        ma2 = params['ma2']
        simple_ma = params['simple_ma']

        if simple_ma:

            # Create short and long simple moving averages
            ma_1 = np.array(prices['Close'].rolling(ma1).mean())
            ma_2 = np.array(prices['Close'].rolling(ma2).mean())

        else:
            # Create short and long exponential moving averages
            ma_1 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma1)
            ma_2 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma2)

        # Create start point from first valid number
        start = np.where(~np.isnan(ma_2))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(ma_2))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(ma_2))

        # Create numpy arrays of zeros to store min and max ma values
        min_ma = np.array([0]*len(ma_2))
        max_ma = np.array([0]*len(ma_2))

        # for each row in the DataFrame after the longest MA has started
        for row in range(start, len(ma_2)):

            # Calculate the min and max ma values
            min_ma[row] = min(ma_1[row], ma_2[row], prices['Close'][row])
            max_ma[row] = max(ma_1[row], ma_2[row], prices['Close'][row])

            # If the short MA crosses above the long MA
            if ma_1[row] > ma_2[row] and ma_1[row-1] < ma_2[row-1]:

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the short MA crosses below the long MA
            elif ma_1[row] < ma_2[row] and ma_1[row-1] > ma_2[row-1]:

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                position_signal[row] = position_signal[row-1]
                trade_signal[row] = 0

        prices['ma_1'] = ma_1
        prices['ma_2'] = ma_2
        prices['min_ma'] = min_ma
        prices['max_ma'] = max_ma

        return prices, start, trade_signal


    @staticmethod
    def entry_triple_ma_crossover(prices, params):
        """
        Entry signal for Triple Moving Average Crossover strategy

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        ma1 : Int
            The first moving average. The default is 4.
        ma2 : Int
            The second moving average. The default is 9.
        ma3 : Int
            The third moving average. The default is 18.
        simple_ma : Bool
            Whether to calculate a simple or exponential moving average. The
            default is True.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals.

        """

        ma1 = params['ma1']
        ma2 = params['ma2']
        ma3 = params['ma3']
        simple_ma = params['simple_ma']

        # Create fast, medium and slow simple moving averages
        if simple_ma:
            ma_1 = np.array(prices['Close'].rolling(ma1).mean())
            ma_2 = np.array(prices['Close'].rolling(ma2).mean())
            ma_3 = np.array(prices['Close'].rolling(ma3).mean())

        else:
            ma_1 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma1)
            ma_2 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma2)
            ma_3 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma3)

        # Create start point from first valid number
        start = np.where(~np.isnan(ma_3))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(ma_3))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(ma_3))

        # Create numpy arrays of zeros to store min and max ma values
        min_ma = np.array([0]*len(ma_3))
        max_ma = np.array([0]*len(ma_3))

        # for each row in the DataFrame after the longest MA has started
        for row in range(start, len(ma_3)):

            # Calculate the min and max ma values
            min_ma[row] = min(ma_1[row], ma_2[row], prices['Close'][row])
            max_ma[row] = max(ma_1[row], ma_2[row], prices['Close'][row])

            # If the shortest ma is above the medium ma is above the long ma
            if ma_1[row] > ma_2[row] > ma_3[row]:

                # Set the position signal to long
                position_signal[row] = 1

                # If this was not the case previously
                if not ma_1[row-1] > ma_2[row-1] > ma_3[row-1]:

                    # Signal to go long
                    trade_signal[row] = 1 - position_signal[row-1]

            # If the shortest ma is below the medium ma is below the long ma
            elif ma_1[row] < ma_2[row] < ma_3[row]:

                # Set the position signal to short
                position_signal[row] = -1

                # If this was not the case previously
                if not ma_1[row-1] < ma_2[row-1] < ma_3[row-1]:

                    # Signal to go short
                    trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, the position should be flat
            else:
                position_signal[row] = 0
                trade_signal[row] = 0 - position_signal[row-1]

        # Assign the series to the OHLC data
        prices['ma_1'] = ma_1
        prices['ma_2'] = ma_2
        prices['ma_3'] = ma_3
        prices['min_ma'] = min_ma
        prices['max_ma'] = max_ma

        return prices, start, trade_signal


    @staticmethod
    def entry_quad_ma_crossover(prices, params):

        """
        Entry signals for Quad Moving Average strategy

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        ma1 : Int, optional
            The fastest of the 4 moving averages. The default is 5 periods.
        ma2 : Int, optional
            The 2nd fastest of the 4 moving averages. The default is 12
            periods.
        ma3 : Int, optional
            The second slowest of the 4 moving averages. The default is 20
            periods.
        ma4 : Int, optional
            The slowest of the 4 moving averages. The default is 40 periods.
        simple_ma : Bool
            Whether to calculate a simple or exponential moving average. The
            default is True.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals.
        """

        ma1 = params['ma1']
        ma2 = params['ma2']
        ma3 = params['ma3']
        ma4 = params['ma4']
        simple_ma = params['simple_ma']

        # Create the 4 simple moving averages
        if simple_ma:
            ma_1 = prices['Close'].rolling(ma1).mean()
            ma_2 = prices['Close'].rolling(ma2).mean()
            ma_3 = prices['Close'].rolling(ma3).mean()
            ma_4 = prices['Close'].rolling(ma4).mean()

        else:
            ma_1 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma1)
            ma_2 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma2)
            ma_3 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma3)
            ma_4 = Indicators.EMA(
                input_series=prices['Close'], time_period=ma4)

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(ma_4))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(ma_4))

        # Create start point from first valid number
        start = np.where(~np.isnan(ma_4))[0][0]

        # Create numpy arrays of zeros to store min and max ma values
        min_ma = np.array([0]*len(ma_4))
        max_ma = np.array([0]*len(ma_4))

        # for each row in the DataFrame after the longest MA has started
        for row in range(start + 1, len(ma_4)):

            # Calculate the min and max ma values
            min_ma[row] = min(ma_1[row], ma_2[row], prices['Close'][row])
            max_ma[row] = max(ma_1[row], ma_2[row], prices['Close'][row])

            # If the second slowest MA is above the slowest MA
            if ma_3[row] > ma_4[row]:

                # If the fastest MA crosses above the second fastest MA
                if ma_1[row] > ma_2[row] and ma_1[row - 1] < ma_2[row - 1]:

                    # Set the position signal to long
                    position_signal[row] = 1

                    # Signal to go long
                    trade_signal[row] = 1 - position_signal[row-1]

                # If the fastest MA crosses below the second fastest MA
                elif ma_1[row] < ma_2[row] and ma_1[row - 1] > ma_2[row - 1]:

                    # If there is a position on
                    if position_signal[row-1] != 0:

                        # Set the position signal to flat
                        position_signal[row] = 0

                        # Signal to close out long
                        trade_signal[row] = -1

                # Otherwise, take no action
                else:
                    trade_signal[row] = 0
                    position_signal[row] = position_signal[row-1]


            # If the second slowest MA is below the slowest MA
            else:

                # If the fastest MA crosses below the second fastest MA
                if ma_1[row] < ma_2[row] and ma_1[row - 1] > ma_2[row - 1]:

                    # Set the position signal to short
                    position_signal[row] = -1

                    # Signal to go short
                    trade_signal[row] = -1 - position_signal[row-1]

                # If the fastest MA crosses above the second fastest MA
                elif ma_1[row] > ma_2[row] and ma_1[row - 1] < ma_2[row - 1]:

                    # If there is a position on
                    if position_signal[row-1] != 0:

                        # Set the position to flat
                        position_signal[row] = 0

                        # Signal to close out short
                        trade_signal[row] = 1

                # Otherwise, take no action
                else:
                    trade_signal[row] = 0
                    position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['ma_1'] = ma_1
        prices['ma_2'] = ma_2
        prices['ma_3'] = ma_3
        prices['ma_4'] = ma_4
        prices['min_ma'] = min_ma
        prices['max_ma'] = max_ma

        return prices, start, trade_signal



"""
Calculate trades data

"""

import numpy as np

class Trades():
    """
    Calculate trade numbers, prices and combined trade signals

    """
    @staticmethod
    def trade_numbers(prices, end_of_day_position, start):
        """
        Calculate the trade numbers

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        end_of_day_position : Series
            The number of units of position held at the end of day.
        start : Int
            The first valid row to start calculating trade information from.

        Returns
        -------
        trade_number : Series
            Array of trade numbers.

        """
        # Extract the end of day position from the OHLC Data
        end_of_day_position = np.array(end_of_day_position)

        # Create numpy array of zeros to store trade numbers
        trade_number = np.array([0]*len(prices))

        # Set initial trade count to zero
        trade_count = 0

        # For each valid row in the data
        for row in range(start + 1, len(prices)):

            # If today's position is zero
            if end_of_day_position[row] == 0:

                # If yesterday's position is zero
                if end_of_day_position[row - 1] == 0:

                    # There is no open trade so set trade number to zero
                    trade_number[row] = 0

                # If yesterday's position is not zero
                else:

                    # Set the trade number to the current trade count
                    trade_number[row] = trade_count

            # If today's position is the same as yesterday
            elif end_of_day_position[row] == end_of_day_position[row - 1]:

                # Set the trade number to yesterdays trade number
                trade_number[row] = trade_number[row - 1]

            # If today's position is non-zero and different from yesterday
            else:

                # Increase trade count by one for a new trade
                trade_count += 1

                # Set the trade number to the current trade count
                trade_number[row] = trade_count

        return trade_number


    @staticmethod
    def trade_prices(prices, trade_number):
        """
        Calculate per trade entry, exit, high and low prices.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.

        Returns
        -------
        trade_entry_price : Series
            The entry price for each trade.
        trade_exit_price : Series
            The exit price for each trade.
        trade_high_price : Series
            The high price for each trade.
        trade_low_price : Series
            The low price for each trade.
        trade_close_high_price : Series
            The highest closing price for each trade.
        trade_close_low_price : Series
            The lowest closing price for each trade.

        """

        # Initialize price arrays with zeroes
        trade_entry_price = np.array([0.0]*len(prices))
        trade_exit_price = np.array([0.0]*len(prices))
        trade_high_price = np.array([0.0]*len(prices))
        trade_low_price = np.array([0.0]*len(prices))
        trade_close_high_price = np.array([0.0]*len(prices))
        trade_close_low_price = np.array([0.0]*len(prices))

        # For each row in the DataFrame
        for row in range(len(prices)):

            # Get the current trade number
            trade_num = trade_number[row]

            # Get the index location of the trade entry date
            trade_first_row = prices.index.get_loc(
                prices[trade_number==trade_num].index[0])

            # Get the number of days since trade entry
            trade_row_num = row - trade_first_row

            # If there is no current trade, set the fields to zero
            if trade_num == 0:
                trade_entry_price[row] = 0
                trade_high_price[row] = 0
                trade_low_price[row] = 0

            # Otherwise
            else:
                # Set the trade entry price to the opening price on the first
                # trade date
                trade_entry_price[row] = prices[
                    trade_number==trade_num].iloc[0]['Open']

                # Set the trade exit price to the opening price on the last
                # trade date
                trade_exit_price[row] = prices[
                    trade_number==trade_num].iloc[-1]['Open']

                # Calculate the maximum close price during the trade
                trade_close_high_price[row] = max(
                    prices[trade_number==trade_num].iloc[
                        0:trade_row_num+1]['Close'])

                # Calculate the minimum close price during the trade
                trade_close_low_price[row] = min(
                    prices[trade_number==trade_num].iloc[
                        0:trade_row_num+1]['Close'])

                # Calculate the maximum high price during the trade
                trade_high_price[row] = max(
                    prices[trade_number==trade_num].iloc[
                        0:trade_row_num+1]['High'])

                # Calculate the minimum low price during the trade
                trade_low_price[row] = min(
                    prices[trade_number==trade_num].iloc[
                        0:trade_row_num+1]['Low'])

        # Create dict to store arrays
        trade_price_dict = {}
        trade_price_dict['trade_entry_price'] = trade_entry_price
        trade_price_dict['trade_exit_price'] = trade_exit_price
        trade_price_dict['trade_high_price'] = trade_high_price
        trade_price_dict['trade_low_price'] = trade_low_price
        trade_price_dict['trade_close_high_price'] = trade_close_high_price
        trade_price_dict['trade_close_low_price'] = trade_close_low_price

        return trade_price_dict


    @staticmethod
    def signal_combine(
            prices, start, end_of_day_position, trade_signals):
        """
        Combine Entry, Exit and Stop signals into a single composite signal.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        start : Int
            The first valid row to start calculating trade information from.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        trade_signals : DataFrame
            DataFrame of Entry, exit and stop signals:
                raw_trade_signal : Series
                    The series of raw Buy / Sell entry signals.
                exit_signal : Series
                    The series of Buy / Sell exit signals.
                stop_signal : Series
                    The series of Buy / Sell stop signals.

        Returns
        -------
        combined_signal : Series
            The series of combined Buy / Sell signals.

        """

        # Create an empty array to store the signals
        combined_signal = np.array([0.0]*len(prices))

        raw_trade_signal = trade_signals['raw_trade_signal']
        trade_number = prices['raw_trade_number']

        # Set a flag for whether the exit has not been triggered for each trade
        flag = True

        # For each row in the data
        for row in range(start, len(prices)):

            # Find the row that relates to the trade exit
            trade_num = trade_number[row]
            trade_last_row = prices.index.get_loc(
                prices[trade_number==trade_num].index[-1])

            # If the raw trade signal is to change position and the current
            # end of day position is flat
            if ((raw_trade_signal[row] != 0) and (
                    end_of_day_position[row] == 0)):

                # Set the combined signal to the raw entry signal
                combined_signal[row] = raw_trade_signal[row]

            else:
                # If there is a trade on (based on the raw entry signal)
                if trade_number[row] != 0:

                    # If an exit has not yet been triggered
                    if flag:

                        # If there is a long position
                        if end_of_day_position[row] > 0:

                            # Set the trade signal to the minimum of the three
                            # series i.e. take any exit signal
                            combined_signal[row] = int(
                                min(trade_signals.iloc[row]))

                        # If there is a short position
                        elif end_of_day_position[row] < 0:

                            # Set the trade signal to the maximum of the three
                            # series i.e. take any exit signal
                            combined_signal[row] = int(
                                max(trade_signals.iloc[row]))

                        # If the position is flat
                        else:
                            # Set the trade signal to the raw entry signal
                            combined_signal[row] = raw_trade_signal[row]

                        # If there is an exit signalled by any of the three
                        # series
                        if combined_signal[row] != 0:

                            # Set the exit flag to False
                            flag = False

                    # If an exit has been triggered and the current trade
                    # number is not the same as the previous day
                    elif trade_number[row] != trade_number[row-1]:

                        # Set the trade signal to the raw entry signal
                        combined_signal[row] = raw_trade_signal[row]

                        # Reset the exit flag
                        flag=True

                    # If an exit has been triggered and the current trade
                    # number is the same as the previous day and the raw entry
                    # signal is to change position
                    elif row == trade_last_row and abs(
                            raw_trade_signal[row]) > 1:

                        # Set the trade signal to 1/2 raw entry signal
                        combined_signal[row] = int(
                            raw_trade_signal[row] / 2)

                        # Reset the exit flag
                        flag=True

                    else:
                        # Set the trade signal to 0
                        combined_signal[row] = 0

        return combined_signal



"""
Trade exit triggers

"""

import numpy as np

class TradeTargets():
    """
    Calculate profit targets, stop loss levels etc.

    """
    @classmethod
    def exit_and_stop_targets(
            cls, prices, params, trade_price_dict):
        """
        Calculate exit and stop targets.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        exit_amount : Float
            The dollar exit amount. The default is $1000.00.
        stop_amount : Float
            The dollar stop amount. The default is $500.00.
        position_size : Int, optional
            The number of units to trade. The default is based on equity.
        trade_price_dict : Dict
            Dictionary of trade entry/high/low/close series.

        Returns
        -------
        prices : DataFrame
            The OHLC data.

        """

        prices = cls._exit_targets(
            prices=prices,
            exit_amount=params['exit_amount'],
            trade_price_dict=trade_price_dict,
            params=params)

        prices = cls._stop_targets(
            prices=prices,
            stop_amount=params['stop_amount'],
            trade_price_dict=trade_price_dict,
            params=params)

        return prices


    @classmethod
    def _exit_targets(
            cls, prices, exit_amount, trade_price_dict, params):
        """
        Create 4 series of exit targets

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        exit_amount : Float
            The dollar exit amount. The default is $1000.00.
        position_size : Int, optional
            The number of units to trade. The default is based on equity.
        trade_price_dict : Dict
            Dictionary of trade entry/high/low/close series.

        Returns
        -------
        prices : DataFrame
            The OHLC data..

        """
        # Generate profit targets / trailing stops
        prices['exit_profit_target'], prices['exit_initial_dollar_loss'], \
            prices['exit_trailing_close'], \
                prices['exit_trailing_high_low'] = cls._pnl_targets(
                    prices=prices, dollar_amount=exit_amount,
                    trade_price_dict=trade_price_dict,
                    params=params)

        return prices


    @classmethod
    def _stop_targets(
            cls, prices, stop_amount, trade_price_dict, params):
        """
        Create 4 series of stop targets

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        stop_amount : Float
            The dollar stop amount. The default is $500.00.
        position_size : Int, optional
            The number of units to trade. The default is based on equity.
        trade_price_dict : Dict
            Dictionary of trade entry/high/low/close series.

        Returns
        -------
        prices : DataFrame
            The OHLC data.

        """
        # Generate profit targets / trailing stops
        prices['stop_profit_target'], prices['stop_initial_dollar_loss'], \
            prices['stop_trailing_close'], \
                prices['stop_trailing_high_low'] = cls._pnl_targets(
                    prices=prices, dollar_amount=stop_amount,
                    trade_price_dict=trade_price_dict,
                    params=params)

        return prices


    @staticmethod
    def _pnl_targets(
            prices, dollar_amount, trade_price_dict, params):
        """
        Create profit and loss stop and exit points

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        exit_amount : Float
            The dollar exit amount. The default is $1000.00.
        position_size : Int
            The number of units of the chosen ticker to trade.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The number of units of position held at the end of day.
        trade_entry_price : Series
            The entry price for each trade.
        trade_high_price : Series
            The high price for each trade.
        trade_low_price : Series
            The low price for each trade.
        trade_close_high_price : Series
            The highest closing price for each trade.
        trade_close_low_price : Series
            The lowest closing price for each trade.

        Returns
        -------
        profit_target : Series
            The exit levels for each trade based on a dollar loss from the
            entry level.
        initial_dollar_loss : Series
            The exit levels for each trade based on a profit target
        trailing_close : Series
            The exit levels for each trade based on the trailing close.
        trailing_high_low : Series
            The exit levels for each trade based on the trailing high / low.

        """

        # Create empty arrays to store the values
        trade_target = np.array([0.0]*len(prices))
        profit_target = np.array([0.0]*len(prices))
        initial_dollar_loss = np.array([0.0]*len(prices))
        trailing_close = np.array([0.0]*len(prices))
        trailing_high_low = np.array([0.0]*len(prices))

        end_of_day_position = prices['raw_end_of_day_position']
        position_size = prices['position_size']

        # For each row in the data
        for row in range(1, len(prices)):

            # Calculate the trade target (distance that price has to change) by
            # dividing the dollar amount by the number of units making up the
            # position size
            if position_size[row] == 0:
                trade_target[row] = 0
            else:
                trade_target[row] = np.round(
                    (dollar_amount / params['contract_point_value'])
                    / position_size[row], 2)

            # If there is a trade on
            if prices['raw_trade_number'][row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # Set the profit target to the trade entry price plus the
                    # trade target
                    profit_target[row] = (
                        trade_price_dict['trade_entry_price'][row]
                        + trade_target[row])

                    # Set the initial dollar loss target to the trade entry
                    # price minus the trade target
                    initial_dollar_loss[row] = (
                        trade_price_dict['trade_entry_price'][row]
                        - trade_target[row])

                    # Set the trailing close target to the closing high price
                    # of the trade minus the trade target
                    trailing_close[row] = (
                        trade_price_dict['trade_close_high_price'][row]
                        - trade_target[row])

                    # Set the trailing high/low target to the high price
                    # of the trade minus the trade target
                    trailing_high_low[row] = (
                        trade_price_dict['trade_high_price'][row]
                        - trade_target[row])

                # If there is a short position
                else:
                    # Set the profit target to the trade entry price minus the
                    # trade target
                    profit_target[row] = (
                        trade_price_dict['trade_entry_price'][row]
                        - trade_target[row])

                    # Set the initial dollar loss target to the trade entry
                    # price plus the trade target
                    initial_dollar_loss[row] = (
                        trade_price_dict['trade_entry_price'][row]
                        + trade_target[row])

                    # Set the trailing close target to the closing low price
                    # of the trade plus the trade target
                    trailing_close[row] = (
                        trade_price_dict['trade_close_low_price'][row]
                        + trade_target[row])

                    # Set the trailing high/low target to the low price
                    # of the trade minus the trade target
                    trailing_high_low[row] = (
                        trade_price_dict['trade_low_price'][row]
                        + trade_target[row])

        return profit_target, initial_dollar_loss, trailing_close, \
            trailing_high_low



"""
Graph the performance of the trading strategy

"""

import matplotlib.pyplot as plt

class PerformanceGraph():
    """
    Class of functions used to graph trading system performance

    """

    @classmethod
    def two_panel_graph(
            cls, signals=None, tables=None, params=None, es_dict=None):
        """
        Create the 2 panel graph

        Parameters
        ----------
        signals : Bool
            Whether to plot the Buy/Sell signals on the price chart.

        Returns
        -------
        Matplotlib Chart
            Displays the chart.

        """

        # Setup initialization variables
        graph_params = cls._graph_variables(
                prices=tables['prices'], entry_type=params['entry_type'],
                entry_signal_indicators=es_dict['entry_signal_indicators'])

        # Set the matplotlib style to use for the charts
        with plt.style.context('fivethirtyeight'):

            # Set up the 2 graphs
            ax1, ax2 = cls._two_panel_setup(
                prices=tables['prices'], graph_params=graph_params,
                params=params, es_dict=es_dict)

            # If signals are to be plotted
            if signals:

                # Create the trade signal points
                signal_dict = cls._create_signals(
                    prices=tables['prices'], graph_params=graph_params)

                # Add these to the price chart
                ax1 = cls._plot_signals(axis=ax1, signal_dict=signal_dict)

            # Add legend, labels and titles to the graphs
            ax1, ax2 = cls._two_panel_legend(
                ax1=ax1, ax2=ax2, perf_dict=tables['perf_dict'])

        # Plot the graphs
        plt.show()


    @classmethod
    def three_panel_graph(
            cls, signals=None, tables=None, params=None, es_dict=None):
        """
        Create the 3 panel graph

        Parameters
        ----------
        signals : Bool
            Whether to plot the Buy/Sell signals on the price chart.

        Returns
        -------
        Matplotlib Chart
            Displays the chart.

        """

        # Setup initialization variables
        graph_params = cls._graph_variables(
                prices=tables['prices'], entry_type=params['entry_type'],
                entry_signal_indicators=es_dict['entry_signal_indicators'])

        # All but the Stochastic Entry methods use a single indicator column
        if 'stoch' not in params['entry_type']:
            indicator = tables['prices'][
                es_dict['entry_signal_indicators'][params['entry_type']]]
        else:
            indicator = None

        # Set the matplotlib style to use for the charts
        with plt.style.context('fivethirtyeight'):

            # Set up the 3 graphs
            ax1, ax2, ax3 = cls._three_panel_setup(
                prices=tables['prices'], graph_params=graph_params,
                params=params, es_dict=es_dict)

            # If signals are to be plotted
            if signals:

                # Create the trade signal points
                signal_dict = cls._create_signals(
                    prices=tables['prices'], graph_params=graph_params)

                # Add these to the price chart
                ax1 = cls._plot_signals(axis=ax1, signal_dict=signal_dict)

            # Set the overbought / oversold lines on the indicator chart
            ax2 = cls._indicator_format(
                axis=ax2, dates=graph_params['dates'], indicator=indicator,
                params=params)

            # Add legend, labels and titles to the graphs
            axes = {'ax1':ax1, 'ax2':ax2, 'ax3':ax3}
            ax1, ax2, ax3 = cls._three_panel_legend(
                axes=axes, perf_dict=tables['perf_dict'],
                params=params,
                entry_signal_labels=es_dict['entry_signal_labels'])

        # Plot the graphs
        plt.show()


    @staticmethod
    def _graph_variables(
            prices=None, entry_type=None, entry_signal_indicators=None):
        """
        Create graph initialization variables

        Returns
        -------
        dates : Pandas Series
            The dates to plot on the x-axis.
        price : Pandas Series
            Closing Prices.
        equity : Pandas Series
            Daily Mark to Market Equity level.
        cumsig : Pandas Series
            The cumulative buy / sell signal.
        lower_bound : Pandas Series
            Lower point to set where trading signals are plotted from.
        upper_bound : Pandas Series
            Upper point to set where trading signals are plotted from.

        """

        # Dictionary to store default params
        graph_params = {}

        # Set the dates to the index of the main DataFrame
        graph_params['dates'] = prices.index

        # Closing Prices
        graph_params['price'] = prices['Close']

        # MTM Equity
        graph_params['equity'] = prices['mtm_equity']

        # Cumulative sum of the combined entry, exit and stop signal
        graph_params['cumsig'] = prices['combined_signal'].cumsum()

        # The lower and upper bounds are used in setting where the trade
        # signals are plotted on the price chart
        # If the entry is a channel breakout
        if entry_type == 'channel_breakout':

            # Set the lower bound as rolling low close prices
            graph_params['lower_bound'] = prices[
                entry_signal_indicators[entry_type][0]]

            # Set the upper bound as rolling high close prices
            graph_params['upper_bound'] = prices[
                entry_signal_indicators[entry_type][1]]

        elif entry_type in ['2ma', '3ma', '4ma']:

            # Set the lower bound as the lowest of the moving average values
            # and the price
            graph_params['lower_bound'] = prices['min_ma']

            # Set the upper bound as the highest of the moving average values
            # and the price
            graph_params['upper_bound'] = prices['max_ma']

        # Otherwise
        else:
            # Set the upper and lower bounds to the closing price
            graph_params['lower_bound'] = graph_params['price']
            graph_params['upper_bound'] = graph_params['price']

        return graph_params


    @staticmethod
    def _two_panel_setup(
            prices=None, graph_params=None, params=None, es_dict=None):
        """
        Set up the 2 panel chart

        Parameters
        ----------
        dates : Pandas Series
            The dates to plot on the x-axis.
        price : Pandas Series
            Closing Prices.
        equity : Pandas Series
            Daily Mark to Market Equity level.

        Returns
        -------
        ax1 : Matplotlib Axes object
            The main price chart.
        ax2 : Matplotlib Axes object
            The MTM Equity chart.

        """

        # If the entry is a moving average
        if params['entry_type'] in ['2ma', '3ma', '4ma']:

            # The indicator label has 2 parts to combine
            indicator_label = (
                es_dict['entry_signal_labels'][params['entry_type']][0]
                + es_dict['entry_signal_labels'][params['entry_type']][1])
        # Otherwise
        else:
            # Assign the label from the default dictionary
            indicator_label = es_dict[
                'entry_signal_labels'][params['entry_type']]

        # Set the figure size to 25 wide and 10 high
        plt.rcParams['figure.figsize'] = (25, 10)

        # Plot the 2 charts as subplots, the first 5 rows high, starting at
        # zero, the second 3 rows high, starting at 6
        ax1 = plt.subplot2grid((9,1), (0,0), rowspan = 5, colspan = 1)
        ax2 = plt.subplot2grid((9,1), (6,0), rowspan = 3, colspan = 1)

        # Plot price against time in the first graph
        ax1.plot(graph_params['dates'],
                 graph_params['price'],
                 linewidth=1.5,
                 label='Close Price')

        # If the entry is Parabolic SAR
        if params['entry_type'] == 'sar':

            # Extract the SAR series from the core DataFrame
            indicator = prices[
                es_dict['entry_signal_indicators'][params['entry_type']]]

            # Add these as a scatter to the price-time chart
            ax1.scatter(graph_params['dates'],
                        indicator,
                        marker='.',
                        color='red',
                        s=10,
                        label=str(params['entry_period'])
                        +'-day '
                        +indicator_label)

        # If the entry is a channel breakout
        elif params['entry_type'] == 'channel_breakout':

            # Extract Rolling Low and Rolling High Closes
            lower_channel = prices[
                es_dict['entry_signal_indicators'][params['entry_type']][0]]
            upper_channel = prices[
                es_dict['entry_signal_indicators'][params['entry_type']][1]]

            # Add these to the price-time chart
            ax1.plot(graph_params['dates'],
                     lower_channel,
                     linewidth=2,
                     label=str(params['entry_period'])+'-day Low')
            ax1.plot(graph_params['dates'],
                     upper_channel,
                     linewidth=2,
                     label=str(params['entry_period'])+'-day High')

        # If the entry is a moving average
        elif params['entry_type'] in ['2ma', '3ma', '4ma']:

            # Extract and plot the first 2 moving averages
            ma_1 = prices[
                es_dict['entry_signal_indicators'][params['entry_type']][0]]
            ax1.plot(graph_params['dates'],
                     ma_1,
                     linewidth=2,
                     label=str(params['ma1'])+'-day MA')

            ma_2 = prices[
                es_dict['entry_signal_indicators'][params['entry_type']][1]]
            ax1.plot(graph_params['dates'],
                     ma_2,
                     linewidth=2,
                     label=str(params['ma2'])+'-day MA')

            # If the entry type is not a double moving average
            if params['entry_type'] != '2ma':

                # Plot the third moving average
                ma_3 = prices[
                    es_dict['entry_signal_indicators'][
                        params['entry_type']][2]]
                ax1.plot(graph_params['dates'],
                         ma_3,
                         linewidth=2,
                         label=str(params['ma3'])+'-day MA')

                # If the entry type is not the triple moving average
                if params['entry_type'] != '3ma':

                    # Plot the fourth moving average
                    ma_4 = prices[
                        es_dict['entry_signal_indicators'][
                            params['entry_type']][3]]
                    ax1.plot(graph_params['dates'],
                             ma_4,
                             linewidth=2,
                             label=str(params['ma4'])+'-day MA')

        # Plot the MTM Equity curve
        ax2.plot(graph_params['dates'],
                 graph_params['equity'],
                 linewidth=1.5,
                 label='MTM Equity')

        return ax1, ax2


    @staticmethod
    def _three_panel_setup(
            prices=None, graph_params=None, params=None, es_dict=None):
        """
        Set up the 3 panel chart

        Parameters
        ----------
        dates : Pandas Series
            The dates to plot on the x-axis.
        price : Pandas Series
            Closing Prices.
        equity : Pandas Series
            Daily Mark to Market Equity level.

        Returns
        -------
        ax1 : Matplotlib Axes object
            The main price chart.
        ax2 : Matplotlib Axes object
            The Indicator chart.
        ax3 : Matplotlib Axes object
            The MTM Equity chart.

        """

         # Assign the label from the default dictionary
        indicator_label = es_dict[
                'entry_signal_labels'][params['entry_type']]

        # Set the figure size to 25 wide and 14 high
        plt.rcParams['figure.figsize'] = (25, 14)

        # Plot the 3 charts as subplots, the first 5 rows high, starting at
        # zero, the second 3 rows high, starting at 6, the third 3 rows high
        # starting at 10
        ax1 = plt.subplot2grid((13,1), (0,0), rowspan = 5, colspan = 1)
        ax2 = plt.subplot2grid((13,1), (6,0), rowspan = 3, colspan = 1)
        ax3 = plt.subplot2grid((13,1), (10,0), rowspan = 3, colspan = 1)

        # Plot price against time in the first graph
        ax1.plot(graph_params['dates'],
                 graph_params['price'],
                 color='black',
                 linewidth=1.5,
                 label='Close Price')

        # If the trade entry is a Stochastic
        if 'stoch' in params['entry_type']:

            # Take the slow k and d values from the core DataFrame
            slow_k = prices[
                es_dict['entry_signal_indicators'][params['entry_type']][0]]
            slow_d = prices[
                es_dict['entry_signal_indicators'][params['entry_type']][1]]

            # Plot these in the second chart
            ax2.plot(graph_params['dates'],
                     slow_k,
                     color='blue',
                     linewidth=1.5,
                     label='Slow k')
            ax2.plot(graph_params['dates'],
                     slow_d,
                     color='red',
                     linewidth=1.5,
                     label='Slow d')

        # Otherwise
        else:
            # Take the indicator column based on entry type and plot in the
            # second chart
            indicator = prices[
                es_dict['entry_signal_indicators'][params['entry_type']]]
            ax2.plot(graph_params['dates'],
                     indicator,
                     color='blue',
                     linewidth=1.5,
                     label=str(params['entry_period'])+'-day '+indicator_label)

        # Plot the MTM Equity curve
        ax3.plot(graph_params['dates'],
                 graph_params['equity'],
                 linewidth=1.5,
                 label='MTM Equity')

        return ax1, ax2, ax3


    @staticmethod
    def _create_signals(
            prices=None, graph_params=None):
        """
        Create trade signals to be plotted on main price chart

        Parameters
        ----------
        cumsig : Pandas Series
            The cumulative buy / sell signal.
        lower_bound : Pandas Series
            Lower point to set where trading signals are plotted from.
        upper_bound : Pandas Series
            Upper point to set where trading signals are plotted from.

        Returns
        -------
        signal_dict : Dict
            Dictionary containing the trade signal details.

        """
        # Create empty dictionary
        signal_dict = {}
        buy_sell_distance = 0.07
        flat_distance = 0.1

        # Buy signal to go long is where the current cumulative signal is to be
        # long when yesterday it was flat
        signal_dict['buy_long_signals'] = (
            (graph_params['cumsig'] == 1)
            & (graph_params['cumsig'].shift() != 1))

        # Place the marker at the specified distance below the lower bound
        signal_dict['buy_long_marker'] = (
            graph_params['lower_bound']
            * signal_dict['buy_long_signals']
            - graph_params['lower_bound'].max()*buy_sell_distance)

        signal_dict['buy_long_marker'] = signal_dict[
            'buy_long_marker'][signal_dict['buy_long_signals']]

        # Set the dates for the buy long signals
        signal_dict['buy_long_dates'] = prices.index[
            signal_dict['buy_long_signals']]

        # Buy signal to go flat is where the current cumulative signal is to be
        # flat when yesterday it was short
        signal_dict['buy_flat_signals'] = (
            (graph_params['cumsig'] == 0)
            & (graph_params['cumsig'].shift() == -1))

        # Place the marker at the specified distance below the lower bound
        signal_dict['buy_flat_marker'] = (
            graph_params['lower_bound']
            * signal_dict['buy_flat_signals']
            - graph_params['lower_bound'].max()*flat_distance)

        signal_dict['buy_flat_marker'] = signal_dict[
            'buy_flat_marker'][signal_dict['buy_flat_signals']]

        # Set the dates for the buy flat signals
        signal_dict['buy_flat_dates'] = prices.index[
            signal_dict['buy_flat_signals']]

        # Sell signal to go flat is where the current cumulative signal is to
        # be flat when yesterday it was long
        signal_dict['sell_flat_signals'] = (
            (graph_params['cumsig'] == 0)
            & (graph_params['cumsig'].shift() == 1))

        # Place the marker at the specified distance above the upper bound
        signal_dict['sell_flat_marker'] = (
            graph_params['upper_bound']
            * signal_dict['sell_flat_signals']
            + graph_params['upper_bound'].max()*flat_distance)

        signal_dict['sell_flat_marker'] = signal_dict[
            'sell_flat_marker'][signal_dict['sell_flat_signals']]

        # Set the dates for the sell flat signals
        signal_dict['sell_flat_dates'] = prices.index[
            signal_dict['sell_flat_signals']]

        # Set the dates for the sell short signals
        signal_dict['sell_short_signals'] = (
            (graph_params['cumsig'] == -1)
            & (graph_params['cumsig'].shift() != -1))

        # Place the marker at the specified distance above the upper bound
        signal_dict['sell_short_marker'] = (
            graph_params['upper_bound']
            * signal_dict['sell_short_signals']
            + graph_params['upper_bound'].max()*buy_sell_distance)

        signal_dict['sell_short_marker'] = signal_dict[
            'sell_short_marker'][signal_dict['sell_short_signals']]

        # Set the dates for the sell short signals
        signal_dict['sell_short_dates'] = prices.index[
            signal_dict['sell_short_signals']]

        return signal_dict


    @staticmethod
    def _plot_signals(axis=None, signal_dict=None):
        """
        Plot trade signals on price-time chart

        Parameters
        ----------
        axis : Matplotlib Axes object
            The main price chart.
        signal_dict : Dict
            Dictionary containing the trade signal details.

        Returns
        -------
        axis : Matplotlib Axes object
            The main price chart.

        """
        # Plot the signals to buy and go long
        axis.scatter(signal_dict['buy_long_dates'],
                    signal_dict['buy_long_marker'],
                    marker='^',
                    color='green',
                    s=100,
                    label='Go Long')

        # Plot the signals to buy and flatten the position
        axis.scatter(signal_dict['buy_flat_dates'],
                    signal_dict['buy_flat_marker'],
                    marker='^',
                    color='deepskyblue',
                    s=100,
                    label='Flatten Short')

        # Plot the signals to sell and flatten the position
        axis.scatter(signal_dict['sell_flat_dates'],
                    signal_dict['sell_flat_marker'],
                    marker='v',
                    color='teal',
                    s=100,
                    label='Flatten Long')

        # Plot the signals to sell and go short
        axis.scatter(signal_dict['sell_short_dates'],
                    signal_dict['sell_short_marker'],
                    marker='v',
                    color='red',
                    s=100,
                    label='Go Short')

        return axis


    @staticmethod
    def _indicator_format(
            axis=None, indicator=None, dates=None,  params=None):
        """
        Apply Overbought / Oversold formatting to the indicator chart

        Parameters
        ----------
        ax : Matplotlib Axes object
            The Indicator chart.
        dates : Pandas Series
            The dates to plot on the x-axis.
        indicator : Pandas Series
            The values of the indicator for these dates.

        Returns
        -------
        ax : Matplotlib Axes object
            The Indicator chart.

        """

        # For all the indicators other than momentum and volatility
        if params['entry_type'] not in ['momentum', 'volatility']:

            # Plot horizontal overbought and oversold lines
            axis.axhline(
                y=params['entry_oversold'],
                color='black',
                linewidth=1)
            axis.axhline(
                y=params['entry_overbought'],
                color='black',
                linewidth=1)

            # For the RSI and CCI indicators
            if params['entry_type'] in ['rsi', 'cci']:

                # Fill the area over the overbought line to the indicator
                axis.fill_between(
                    dates, indicator, params['entry_overbought'],
                    where=indicator>=params['entry_overbought'],
                    interpolate=True, color='red')

                # Fill the area under the oversold line to the indicator
                axis.fill_between(
                    dates, indicator, params['entry_oversold'],
                    where=indicator<=params['entry_oversold'],
                    interpolate=True, color='red')

        return axis


    @staticmethod
    def _two_panel_legend(ax1, ax2, perf_dict):
        """
        Create the legend, axis labels and titles for the 2 panel graph

        Parameters
        ----------
        ax1 : Matplotlib Axes object
            The main price chart.
        ax2 : Matplotlib Axes object
            The MTM Equity chart.

        Returns
        -------
        ax1 : Matplotlib Axes object
            The main price chart.
        ax2 : Matplotlib Axes object
            The MTM Equity chart.

        """

        # Title of main price chart is the contract longname plus the entry
        # strategy
        ax1.set_title(perf_dict['longname']+' - '+perf_dict['entry_label'])
        ax2.set_title("Equity Curve")
        ax1.set_ylabel('Price')
        ax2.set_ylabel('Equity')
        ax1.legend()

        return ax1, ax2


    @staticmethod
    def _three_panel_legend(
            axes, perf_dict, entry_signal_labels, params):
        """
        Create the legend, axis labels and titles for the 3 panel graph

        Parameters
        ----------
        ax1 : Matplotlib Axes object
            The main price chart.
        ax2 : Matplotlib Axes object
            The Indicator chart.
        ax3 : Matplotlib Axes object
            The MTM Equity chart.

        Returns
        -------
        ax1 : Matplotlib Axes object
            The main price chart.
        ax2 : Matplotlib Axes object
            The Indicator chart.
        ax3 : Matplotlib Axes object
            The MTM Equity chart.

        """
        ax1 = axes['ax1']
        ax2 = axes['ax2']
        ax3 = axes['ax3']

        # Title of main price chart is the contract longname plus the entry
        # strategy
        ax1.set_title(perf_dict['longname']+' - '+perf_dict['entry_label'])
        ax1.set_ylabel('Price')

        # Title of the Indicator chart is the time period plus indicator name
        ax2.set_title(str(params['entry_period'])
                      +'-day '
                      +entry_signal_labels[params['entry_type']])
        ax3.set_title("Equity Curve")
        ax3.set_ylabel('Equity')
        ax1.legend()

        return ax1, ax2, ax3



"""
Market Data functions
"""
import os
import norgatedata
import numpy as np
import pandas as pd
import requests
from yahoofinancials import YahooFinancials

class Markets():
    """
    Methods for collecting data from Norgate Data, Yahoo Finance and
    AlphaVantage and extracting the long names of the Norgate Data tickers.

    """
    @classmethod
    def create_base_data(
            cls, ticker=None, source=None, params=None, bench_flag=None):
        """
        Create DataFrame of OHLC prices from NorgateData or Yahoo Finance

        Parameters
        ----------
        ticker : Str, optional
            Underlying to return. The default '$SPX'.
        ccy_1 : Str, optional
            Primary currency of pair to return. The default 'GBP'.
        ccy_2 : Str, optional
            Secondary currency of pair to return. The default 'USD'.
        start_date : Str, optional
            Date to begin backtest. Format is 'YYYY-MM-DD'.
        end_date : Str, optional
            Date to end backtest. Format is 'YYYY-MM-DD'.
        source : Str, optional
            The data source to use, either 'norgate' or 'yahoo'. The default
            is 'norgate'.
        api_key : Str
            AlphaVantage API key. If not provided will look for
            'ALPHAVANTAGE_API_KEY' in the environment variables.

        Returns
        -------
        prices : DataFrame
            Returns OHLC DataFrame.

        """

        # If a valid source has been provided
        #try:
        # Extract data from Norgate
        if source == 'norgate':
            prices, params = cls._return_norgate_data(
                ticker=ticker, params=params)

        # Extract data from Yahoo Finance
        elif source == 'yahoo':
            prices = cls._return_yahoo_data(
                ticker=ticker,
                start_date=params['start_date'],
                end_date=params['end_date'])

        # Extract data from AlphaVantage
        elif source == 'alpha':
            prices = cls._return_alphavantage_data(
                ticker=ticker, params=params)

        else:
            raise ValueError(
                'Select a data source from yahoo, norgate or alpha')

        if bench_flag is False:
            params = cls._contract_data(
                ticker=ticker, prices=prices, params=params)

        return prices, params

        # Otherwise return error message
        #except ValueError:
        #    return 'Select a data source from yahoo, norgate or alpha'


    @staticmethod
    def _return_norgate_data(ticker, params):

        timeseriesformat = 'pandas-dataframe'
        prices = norgatedata.price_timeseries(
            symbol=ticker,
            start_date=params['start_date'],
            end_date=params['end_date'],
            format=timeseriesformat)

        return prices, params


    @staticmethod
    def _return_yahoo_data(ticker=None, start_date=None, end_date=None):
        """
        Create DataFrame of historic prices for specified ticker using Yahoo
        Finance as the source.

        Parameters
        ----------
        ticker : Int
            Stock to be returned in the form of Reuters RIC code as a
            string.
        start_date : Str, optional
            Date to begin backtest. Format is 'YYYY-MM-DD'.
        end_date : Str, optional
            Date to end backtest. Format is 'YYYY-MM-DD'.
        freq : Int
            Frequency of data - set to 'daily'.

        Returns
        -------
        prices : DataFrame
            DataFrame of historic prices for given ticker.

        """

        # Initialise data class
        yahoo_financials = YahooFinancials(ticker)
        freq='daily'

        # Extract historic prices
        prices = yahoo_financials.get_historical_price_data(
            start_date, end_date, freq)

        # Reformat columns
        prices = pd.DataFrame(
            prices[ticker]['prices']).drop(['date'], axis=1) \
                .rename(columns={'formatted_date':'Date',
                                 'open': 'Open',
                                 'high': 'High',
                                 'low': 'Low',
                                 'close': 'Close',
                                 'volume': 'Volume'}) \
                .loc[:, ['Date','Open','High','Low','Close','Volume']] \
                .set_index('Date')

        # Set Index to Datetime
        prices.index = pd.to_datetime(prices.index)

        return prices


    @classmethod
    def _return_alphavantage_data(
            cls, ticker=None, params=None):
        """
        Create DataFrame of historic prices for specified ticker using
        AlphaVantage as the source.

        Parameters
        ----------
        ticker : Str
            Underlying to return. The default '$SPX'.
        ccy_1 : Str
            Primary currency of pair to return. The default 'GBP'.
        ccy_2 : Str
            Secondary currency of pair to return. The default 'USD'.
        asset_type : Str
            The alphavantage asset class type. The default is 'fx'.
        start_date : Str, optional
            Date to begin backtest. Format is 'YYYY-MM-DD'.
        end_date : Str, optional
            Date to end backtest. Format is 'YYYY-MM-DD'.
        api_key : Str
            AlphaVantage API key. If not provided will look for
            'ALPHAVANTAGE_API_KEY' in the environment variables.

        Returns
        -------
        prices : DataFrame
            DataFrame of historic prices for given ticker.

        """

        # Set API key
        if params['api_key'] == '':
            params['api_key'] = os.getenv('ALPHAVANTAGE_API_KEY')

        # FX pair
        if params['asset_type'] == 'fx':
            prices = cls._alphavantage_fx(
                ccy_1=params['ccy_1'],
                ccy_2=params['ccy_2'],
                api_key=params['api_key'])

        # Cryptocurrency
        elif params['asset_type'] == 'crypto':
            prices = cls._alphavantage_crypto(
                ccy_1=params['ccy_1'],
                ccy_2=params['ccy_2'],
                api_key=params['api_key'])

        # Equity Single stock or Index
        elif params['asset_type'] == 'equity':
            prices = cls._alphavantage_equity(
                ticker=ticker, api_key=params['api_key'])

        # Otherwise raise an error
        else:
            raise ValueError("Please enter a valid asset type")

        # Set Index to Datetime
        prices.index = pd.to_datetime(prices.index)

        # Sort data in ascending order
        prices = prices[::-1]

        # If a start date has been provided
        if params['start_date'] is not None:

            # Set the start variable to this, converting to datetime format
            start = pd.to_datetime(params['start_date'])

        # If no start date is provided
        else:
            # Set the start variable to the first row in the DataFrame
            start = prices.index[0]

        # If an end date has been provided
        if params['end_date'] is not None:

            # Set the end variable to this, converting to datetime format
            end = pd.to_datetime(params['end_date'])

        # If no end date is provided
        else:

            # Set the end variable to the last row in the DataFrame
            end = prices.index[-1]

        # Trim data to specified dates
        prices = prices.loc[start:end]

        return prices


    @staticmethod
    def _alphavantage_fx(ccy_1, ccy_2, api_key):
        """
        Create DataFrame of historic prices for an fx pair using
        AlphaVantage as the source.

        Parameters
        ----------
        ccy_1 : Str
            Primary currency of pair to return. The default 'GBP'.
        ccy_2 : Str
            Secondary currency of pair to return. The default 'USD'.
        api_key : Str
            AlphaVantage API key. If not provided will look for
            'ALPHAVANTAGE_API_KEY' in the environment variables.

        Returns
        -------
        prices :  DataFrame
            DataFrame of historic prices for given ticker.

        """

        # Set url to extract prices from
        base_url = 'https://www.alphavantage.co/query?'

        # Set fx params
        params = {'function': 'FX_DAILY',
                  'from_symbol': ccy_1,
                  'to_symbol': ccy_2,
                  'outputsize':'full',
                  'apikey': api_key}

        response = requests.get(base_url, params=params)
        response_dict = response.json()

        _, header = response.json()

        #Convert to pandas dataframe
        prices = pd.DataFrame.from_dict(
            response_dict[header], orient='index')

        #Clean up column names
        df_cols = [i.split(' ')[1].title() for i in prices.columns]
        prices.columns = df_cols

        # Set datatype to float
        prices = prices.astype(float)

        return prices


    @staticmethod
    def _alphavantage_crypto(ccy_1, ccy_2, api_key):
        """
        Create DataFrame of historic prices for a cryptocurrency pair using
        AlphaVantage as the source.

        Parameters
        ----------
        ccy_1 : Str
            Primary currency of pair to return.
        ccy_2 : Str
            Secondary currency of pair to return.
        api_key : Str
            AlphaVantage API key. If not provided will look for
            'ALPHAVANTAGE_API_KEY' in the environment variables.

        Returns
        -------
        prices :  DataFrame
            DataFrame of historic prices for given ticker.

        """
        # Set url to extract prices from
        base_url = 'https://www.alphavantage.co/query?'

        # Set crypto params
        params = {'function': 'DIGITAL_CURRENCY_DAILY',
                  'symbol': ccy_1,
                  'market': ccy_2,
                  'apikey': api_key}

        response = requests.get(base_url, params=params)
        response_dict = response.json()

        _, header = response.json()

        #Convert to pandas dataframe
        prices = pd.DataFrame.from_dict(
            response_dict[header], orient='index')

        # Select the USD OHLC columns
        prices = prices[
            [prices.columns[1], prices.columns[3], prices.columns[5],
             prices.columns[7]]]

        # Set column names
        prices.columns = ['Open', 'High', 'Low', 'Close']

        # Set datatype to float
        prices = prices.astype(float)

        return prices


    @staticmethod
    def _alphavantage_equity(ticker, api_key):
        """
        Create DataFrame of historic prices for an equity ticker using
        AlphaVantage as the source.

        Parameters
        ----------
        ticker : Str
            Underlying to return. The default '$SPX'.
        api_key : Str
            AlphaVantage API key. If not provided will look for
            'ALPHAVANTAGE_API_KEY' in the environment variables.

        Returns
        -------
        prices :  DataFrame
            DataFrame of historic prices for given ticker.

        """
        # Set url to extract prices from
        base_url = 'https://www.alphavantage.co/query?'

        # Set equity params
        params = {'function': 'TIME_SERIES_DAILY_ADJUSTED',
                  'symbol': ticker,
                  'outputsize':'full',
                  'apikey': api_key}

        response = requests.get(base_url, params=params)
        response_dict = response.json()

        _, header = response.json()

        #Convert to pandas dataframe
        prices = pd.DataFrame.from_dict(
            response_dict[header], orient='index')

        #Clean up column names
        df_cols = [i.split(' ')[1].title() for i in prices.columns]
        prices.columns = df_cols

        # Set datatype to float
        prices = prices.astype(float)

        # Calculate stock split multiplier
        prices['split_mult'] = np.array([1.0]*len(prices))
        for row in range(1, len(prices)):
            if prices['Split'][row] == 1:
                prices['split_mult'][row] = prices['split_mult'][row-1]
            else:
                prices['split_mult'][row] = (prices['split_mult'][row-1]
                                         * prices['Split'][row])

        # Adjust OHLC prices for splits
        prices['O'] = np.round(prices['Open'] / prices['split_mult'], 2)
        prices['H'] = np.round(prices['High'] / prices['split_mult'], 2)
        prices['L'] = np.round(prices['Low'] / prices['split_mult'], 2)
        prices['C'] = np.round(prices['Close'] / prices['split_mult'], 2)

        # Select only OHLC columns
        prices = prices[['O', 'H', 'L', 'C']]

        # Set column names
        prices.columns = ['Open', 'High', 'Low', 'Close']

        return prices


    @staticmethod
    def _contract_data(ticker, prices, params):

        if ticker[0] == '&':
            params['front_ticker'] = (
                ticker[1:]
                +'-'
                +str(prices['Delivery Month'][-1])[:4]
                +params['contract_months'][
                    str(prices['Delivery Month'][-1])[4:6]])

            params['per_contract_margin'] = norgatedata.margin(
                params['front_ticker'])
            params['contract_point_value'] = norgatedata.point_value(
                params['front_ticker'])

        else:
            params['contract_point_value'] = 1

        return params


    @staticmethod
    def norgate_name_dict():
        """
        Create a dictionary of the long names of the Norgate tickers.

        Returns
        -------
        norgate_name_dict : Dict
            Dictionary lookup of Norgate tickers to long names.

        """

        # Get list of the available databases
        alldatabasenames = norgatedata.databases()

        # Create empty dictionary to store names
        norgate_name_dict = {}

        # For each of the available databases
        for database in alldatabasenames:

            # Extract dictionary of dictionaries, one for each ticker in the
            # database
            databasecontents = norgatedata.database(database)

            # For each ticker in the database
            for dicto in databasecontents:

                # Set the key-value pair of the new dictionary to the ticker
                # and long name respectively
                key = dicto['symbol']
                value = dicto['securityname']

                # Whether to include backadjusted / regular continuous futures
                if database == 'Continuous Futures':

                    #if '_CCB' in key:
                    norgate_name_dict[key] = value

                # Don't include the individual futures contracts
                elif database == 'Futures':
                    pass

                # Store the values in the dictionary
                else:
                    norgate_name_dict[key] = value

        return norgate_name_dict



"""
Utility functions

"""

import datetime as dt
import numpy as np
import pandas as pd
from pandas.tseries.offsets import BDay

class Labels():
    """
    Create labels for the Entry, Exit and Stop Strategies.

    """

    @classmethod
    def strategy_labels(cls, params=None, default_dict=None):
        """
        Create label and price signal for chosen strategy

        Parameters
        ----------
        **kwargs : Various
            The input parameters necessary for the chosen strategy.

        Returns
        -------
        entry_label : Str
            The longname of the entry strategy.
        exit_label : Str
            The longname of the exit strategy.
        stop_label : Str
            The longname of the stop strategy.

        """

        entry_label = cls._entry_label(
            params=params, default_dict=default_dict)

        exit_label = cls._exit_label(
            params=params, default_dict=default_dict)

        stop_label = cls._stop_label(
            params=params, default_dict=default_dict)

        return entry_label, exit_label, stop_label


    @staticmethod
    def _entry_label(params=None, default_dict=None):

        # Simple or Exponential Moving Average label
        if params['simple_ma']:
            ma_type_label = 'S'
        else:
            ma_type_label = 'E'


        # Entry labels

        # Double Moving Average Crossover
        if params['entry_type'] == '2ma':

            # Set the entry label
            entry_label = (str(params['ma1'])
                           +'-, '
                           +str(params['ma2'])
                           +'-day : '
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']][0]
                           +ma_type_label
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']][1])


        # Triple Moving Average Crossover
        elif params['entry_type'] == '3ma':

            # Set the entry label
            entry_label = (str(params['ma1'])
                           +'-, '
                           +str(params['ma2'])
                           +'-, '
                           +str(params['ma3'])
                           +'-day : '
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']][0]
                           +ma_type_label
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']][1])


        # Quad Moving Average Crossover
        elif params['entry_type'] == '4ma':

            # Set the entry label
            entry_label = (str(params['ma1'])
                           +'-, '
                           +str(params['ma2'])
                           +'-, '
                           +str(params['ma3'])
                           +'-, '
                           +str(params['ma4'])
                           +'-day : '
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']][0]
                           +ma_type_label
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']][1])


        # Parabolic SAR
        elif params['entry_type'] == 'sar':

            # Set the entry label
            entry_label = (str(params['entry_period'])
                          +'-day '
                          +str(np.round(params[
                              'entry_acceleration_factor'] * 100, 1))
                          +'% AF '
                          +default_dict['df_entry_signal_labels'][
                              params['entry_type']])


        # Channel Breakout
        elif params['entry_type'] == 'channel_breakout':

            # Set the entry label
            entry_label = (str(params['entry_period'])
                           +'-day : '
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']])


        # Stochastic Crossover, Stochastic Pop, Stochastic Over Under and
        # Relative Strength Index
        elif params['entry_type'] in ['stoch_cross', 'stoch_over_under',
                                      'stoch_pop', 'rsi']:

            # Set the entry label
            entry_label = (str(params['entry_period'])
                           +'-day '
                           +str(params['entry_overbought'])
                           +'-'
                           +str(params['entry_oversold'])
                           +' : '
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']])


        # Commodity Channel Index, Momentum and Volatility
        elif params['entry_type'] in ['cci', 'momentum', 'volatility']:

            # Set the entry label
            entry_label = (str(params['entry_period'])
                           +'-day '
                           +str(int(params['entry_threshold']*100))
                           +'% : '
                           +default_dict['df_entry_signal_labels'][
                               params['entry_type']])

        # Otherwise raise an error
        else:
            raise ValueError("Please enter a valid entry type")


        return entry_label


    @staticmethod
    def _exit_label(params=None, default_dict=None):

        # Exit labels

        # Parabolic SAR
        if params['exit_type'] == 'sar':

            # Set the exit label
            exit_label = (str(params['exit_period'])
                          +'-day '
                          +str(np.round(params[
                              'exit_acceleration_factor'] * 100, 1))
                          +'% AF '
                          +default_dict['df_exit_signal_labels'][
                              params['exit_type']])


        # Stochastic Crossover and Trailing Relative Strength Index
        elif params['exit_type'] in ['stoch_cross', 'rsi_trail']:

            # Set the exit label
            exit_label = (str(params['exit_period'])
                           +'-day '
                           +str(params['exit_overbought'])
                           +'-'
                           +str(params['exit_oversold'])
                           +' : '
                           +default_dict['df_exit_signal_labels'][
                               params['exit_type']])


        # Volatility
        elif params['exit_type'] in ['volatility']:

            # Set the exit label
            exit_label = (str(params['exit_period'])
                          +'-day '
                          +str(int(params['exit_threshold']*100))
                          +'% : '
                          +default_dict['df_exit_signal_labels'][
                              params['exit_type']])


        # Trailing Stop and Profit Target
        elif params['exit_type'] in ['trailing_stop', 'profit_target']:

            # Set the exit label
            exit_label = ('$'
                          +str(int(params['exit_amount']))
                          +' '
                          +default_dict['df_exit_signal_labels'][
                               params['exit_type']])


        # Support/Resistance, Key Reversal Day and n-Day Range
        elif params['exit_type'] in ['sup_res', 'key_reversal', 'nday_range']:

            # Set the exit label
            exit_label = (str(params['exit_period'])
                          +'-day '
                          +default_dict['df_exit_signal_labels'][
                              params['exit_type']])

        # Otherwise raise an error
        else:
            raise ValueError("Please enter a valid exit type")

        return exit_label


    @staticmethod
    def _stop_label(params=None, default_dict=None):

        # Stop labels

        # Initial Dollar, Breakeven, Trailing Close and Trailing High Low
        if params['stop_type'] in ['initial_dollar', 'breakeven',
                                   'trail_close', 'trail_high_low']:

            # Set the stop label
            stop_label = ('$'
                          +str(int(params['stop_amount']))
                          +' '
                          +default_dict['df_stop_signal_labels'][
                               params['stop_type']])


        # Support / Resistance and Immediate Profit
        elif params['stop_type'] in ['sup_res', 'immediate_profit']:

            # Set the stop label
            stop_label = (str(params['stop_period'])
                          +'-day '
                          +default_dict['df_stop_signal_labels'][
                              params['stop_type']])


        # Otherwise raise an error
        else:
            raise ValueError("Please enter a valid stop type")


        return stop_label


    @staticmethod
    def position_size_label(params):
        """
        Create position size label

        Parameters
        ----------
        params : Dict
            Dictionary of parameters.

        Returns
        -------
        params : Dict
            Dictionary of parameters.

        """
        if params['position_type'] == 'equity_constant':
            params['position_size_label'] = (
                str(int(params['equity_inv_perc'] * 100))
                +'% Starting Equity / Initial Trade Entry Price')

        elif params['position_type'] == 'equity_variable':
            params['position_size_label'] = (
                str(int(params['equity_inv_perc'] * 100))
                +'% Starting Equity / Trade Entry Price')

        elif params['position_type'] == 'atr':
            params['position_size_label'] = (
                str(params['atr_pos_size'])
                +' day ATR : '
                +str(np.round(params['position_risk_bps'] / 100, 2))
                +'% risk')
        else:
            params['position_size_label'] = (
                str(params['fixed_pos_size'])
                +' contracts fixed size')

        return params


class Dates():
    """
    Date calculation and formatting functions.

    """

    @staticmethod
    def date_set(start_date, end_date, lookback):
        """
        Create start and end dates if not supplied

        Parameters
        ----------
        start_date : Str, optional
            Date to begin backtest. Format is YYYY-MM-DD. The default is 750
            business days prior (circa 3 years).
        end_date : Str, optional
            Date to end backtest. Format is YYYY-MM-DD. The default is the
            last business day.
        lookback : Int, optional
            Number of business days to use for the backtest. The default is 750
            business days (circa 3 years).

        Returns
        -------
        start_date : Str
            Date to begin backtest. Format is YYYY-MM-DD.
        end_date : Str
            Date to end backtest. Format is YYYY-MM-DD.

        """

        # If end date is not provided, set to previous working day
        if end_date is None:
            end_date_as_dt = (dt.datetime.today() - BDay(1)).date()
            end_date = str(end_date_as_dt)

        # If start date is not provided, set to today minus lookback period
        if start_date is None:
            start_date_as_dt = (dt.datetime.today() -
                                pd.Timedelta(days=lookback*(365/250))).date()
            start_date = str(start_date_as_dt)

        return start_date, end_date


class Reformat():
    """
    Functions for mapping / scaling data
    """

    @staticmethod
    def position_scale(pos_dict, position_size):
        """
        Scale raw positions by position size

        Parameters
        ----------
        raw_pos_dict : Dict
            Dictionary of start of day, end of day positions and trade actions.
        position_size : Series
            Array of the position size to be applied each day.

        Returns
        -------
        scaled_dict : Dict
            Dictionary of the 3 arrays, scaled by the position sizes.

        """
        # Create empty dictionary
        scaled_dict = {}

        # For each kew, value combination in the raw position dictionary
        for key, value in pos_dict.items():

            # Multiply the array by the position size array
            scaled_dict[key] = value * position_size

        return scaled_dict


    @staticmethod
    def map_to_prices(prices, input_dict, title_modifier):
        """
        Map dictionary of arrays to the OHLC data

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        input_dict : Dict
            Dictionary of arrays.
        title_modifier : Str
            String to append to the array names.

        Returns
        -------
        prices : DataFrame
            The OHLC data.

        """
        # For each key, value combination in the input dictionary
        for key, value in input_dict.items():

            # Add the array to the OHLC DataFrame appending the title modifier
            # to the beginning of the name
            prices[title_modifier+key] = value

        return prices



"""
Winning and Losing trades and runs

"""

import numpy as np
import pandas as pd

class Runs():
    """
    Calculate data on winniing and losing trades and win/loss runs

    """

    @staticmethod
    def trade_data(prices):
        """
        Create dictionary of trades, count of the number of trades and lists /
        dictionaries of winning and losing trades

        Parameters
        ----------
        prices : DataFrame
            The OHLC data with trade signals.

        Returns
        -------
        trades : Dict
            Dictionary containing all the trades.
        num_trades : Int
            Number of trades.
        trades_win_dict : Dict
            Dictionary of winning trades.
        trades_win_list : TYPE
            List of winning trades.
        trades_loss_dict : TYPE
            Dictionary of losing trades.
        trades_loss_list : TYPE
            List of losing trades.

        """
        # profit per trade
        # Create empty trades dictionary
        trades = {}

        # Count the number of unique trade numbers (less 1 for the 0 start)
        num_trades = len(pd.unique(prices['trade_number'])) - 1

        # For each trade number
        for trade_number in range(1, num_trades+1):

            # Calculate profit as the sum of daily pnl for that trade number
            profit = prices[
                prices['trade_number']==trade_number]['daily_pnl'].sum()

            # Assign this number (rounded to 2dp) to the trades dictionary
            trades[trade_number] = np.round(profit, 2)

        # Split winning and losing trades
        # Create empty win/loss dictionaries and lists
        trades_win_dict = {}
        trades_win_list = list()
        trades_loss_dict = {}
        trades_loss_list = list()

        # for each trade and profit in the trades dictionary
        for key, value in trades.items():

            # If the profit is negative
            if value < 0:

                # Add the trade to the loss dictionary and list (as a tuple)
                trades_loss_dict[key] = value
                trades_loss_list.append((key, value))

            # If the profit is positive
            else:

                # Otherwise add to the win dictionary and list (as a tuple)
                trades_win_dict[key] = value
                trades_win_list.append((key, value))

        trade_data_dict = {
            'trades':trades,
            'num_trades':num_trades,
            'trades_win_dict':trades_win_dict,
            'trades_win_list':trades_win_list,
            'trades_loss_dict':trades_loss_dict,
            'trades_loss_list':trades_loss_list
            }

        return trade_data_dict


    @classmethod
    def trade_runs(cls, input_trades_list, run_type):
        """
        Produce data for winning or losing runs of trades

        Parameters
        ----------
        input_trades_list : List
            List of winning or losing trades.
        run_type : Str, optional
            Whether winning or losing run data is being calculated. The
            default is 'win'.

        Returns
        -------
        max_run_pnl : Float
            Largest profit or loss.
        max_run_count : Int
            Number of trades in largest profit or loss.
        min_run_pnl : Float
            Smallest profit or loss.
        min_run_count : Int
            Number of trades in smallest profit or loss.
        num_runs : Int
            Count of number of winning or losing runs.
        av_run_count : Int
            Count of average number of trades in winning or losing runs.
        av_run_pnl : Float
            Average profit or loss.
        pnl : Tuple
            PNL for each run and number of trades.

        """

        pnl = cls._calc_trade_runs(input_trades_list)

        min_max_run_dict = cls._calc_min_max_runs(pnl, run_type)

        # Count number of runs as the length of the pnl list
        num_runs = len(pnl)

        if pnl:

            # Take the average number of runs as the sum of run lengths in pnl
            # tuple divided by the number of runs
            av_run_count = int(np.round(sum(j for i, j in pnl) / len(pnl), 0))

            # Take the average run pnl as the sum of run pnls in pnl tuple
            # divided by the number of runs
            av_run_pnl = np.round(sum(i for i, j in pnl) / len(pnl), 2)

        else:
            av_run_count = 0
            av_run_pnl = 0

        name_dict = cls._calc_run_names(run_type)

        run_dict = {
            name_dict['max_run_pnl_str']:min_max_run_dict['max_run_pnl'],
            name_dict['max_run_count_str']:min_max_run_dict['max_run_count'],
            name_dict['min_run_pnl_str']:min_max_run_dict['min_run_pnl'],
            name_dict['min_run_count_str']:min_max_run_dict['min_run_count'],
            name_dict['num_runs_str']:num_runs,
            name_dict['av_run_count_str']:av_run_count,
            name_dict['av_run_pnl_str']:av_run_pnl,
            name_dict['pnl_str']:pnl
            }

        return run_dict

    @staticmethod
    def _calc_trade_runs(input_trades_list):

        # Set initial values
        max_run_count = 1
        run_count = 1
        run_trades_list = []
        total_run_trades_list = []
        last_trade_count = 0

        # For each trade in the winning or losing trades list, sorting by the
        # trade number
        for num, trade in enumerate(sorted(input_trades_list)):
            # For the first trade
            if num == 0:

                # Add the trade pnl to the winning / losing trades run list
                run_trades_list.append(trade[1])

                # If this is the last trade
                if num == len(input_trades_list) - 1:
                    total_run_trades_list.append(run_trades_list)

            # Otherwise, if the trade number is next in sequence after the
            # last stored trade number
            elif trade[0] == last_trade_count + 1:

                # Increase the run count by one
                run_count +=1

                # Update the longest run count
                max_run_count = max(max_run_count, run_count)

                # Add the trade pnl to the winning / losing trades run list
                run_trades_list.append(trade[1])

                # If this is the last trade
                if num == len(input_trades_list) - 1:
                    total_run_trades_list.append(run_trades_list)

            # If the trade is not the next in sequence:
            else:

                # Add the current run to the list of all runs
                total_run_trades_list.append(run_trades_list)

                # If this is not the last trade
                if num != len(input_trades_list) - 1:

                    # Reset the winning / losing trades run list
                    run_trades_list = []

                    # Add the trade pnl to the winning / losing trades run list
                    run_trades_list.append(trade[1])

                    # Increase the run count by one
                    run_count = 1

                # If it is the last trade
                else:

                    # Reset the winning / losing trades run list
                    run_trades_list = []

                    # Add the trade pnl to the winning / losing trades run list
                    run_trades_list.append(trade[1])

                    # Add the current run to the list of all runs
                    total_run_trades_list.append(run_trades_list)

            # Set the last trade count number to the current trade number
            last_trade_count = trade[0]

        # Tuple for each run of PNL and number of trades.
        pnl = sorted([(sum(x), len(x)) for x in total_run_trades_list])

        return pnl


    @staticmethod
    def _calc_min_max_runs(pnl, run_type):

        # Values to select for winning runs
        if run_type == 'win':

            # If there are any winning trades
            if pnl:
                max_run_pnl = pnl[-1][0]
                max_run_count = pnl[-1][-1]
                min_run_pnl = pnl[0][0]
                min_run_count = pnl[0][-1]

            # Otherwise set the values to zero
            else:
                max_run_pnl = 0
                max_run_count = 0
                min_run_pnl = 0
                min_run_count = 0

        # Values to select for losing runs
        else:

            # If there are any losing trades
            if pnl:
                max_run_pnl = pnl[0][0]
                max_run_count = pnl[0][-1]
                min_run_pnl = pnl[-1][0]
                min_run_count = pnl[-1][-1]

            # Otherwise set the values to zero
            else:
                max_run_pnl = 0
                max_run_count = 0
                min_run_pnl = 0
                min_run_count = 0

        min_max_run_dict = {
            'max_run_pnl':max_run_pnl,
            'max_run_count':max_run_count,
            'min_run_pnl':min_run_pnl,
            'min_run_count':min_run_count
            }

        return min_max_run_dict


    @staticmethod
    def _calc_run_names(run_type):

        max_run_pnl_str = 'max_'+run_type+'_run_pnl'
        max_run_count_str = 'max_'+run_type+'_run_count'
        min_run_pnl_str = 'min_'+run_type+'_run_pnl'
        min_run_count_str = 'min_'+run_type+'_run_count'
        num_runs_str = 'num_'+run_type+'_runs'
        av_run_count_str = 'av_'+run_type+'_run_count'
        av_run_pnl_str = 'av_'+run_type+'_run_pnl'
        pnl_str = run_type+'_pnl'


        name_dict = {
            'max_run_pnl_str':max_run_pnl_str,
            'max_run_count_str':max_run_count_str,
            'min_run_pnl_str':min_run_pnl_str,
            'min_run_count_str':min_run_count_str,
            'num_runs_str':num_runs_str,
            'av_run_count_str':av_run_count_str,
            'av_run_pnl_str':av_run_pnl_str,
            'pnl_str':pnl_str
            }

        return name_dict



"""
Calculate position data

"""

import math
import numpy as np
from technicalmethods.methods import Indicators
import pandas as pd
pd.options.mode.chained_assignment = None

class Positions():
    """
    Functions for calculating trade and position data.

    """
    @staticmethod
    def calc_positions(prices, signal, start):
        """
        Calculate start of day and end of day positions and any buy / sell
        trade actions

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        signal : Series
            The series of Buy / Sell trade signals.
        start : Int
            The first valid row to start calculating trade information from.
        position_size : Int
            The number of units of the chosen ticker to trade.

        Returns
        -------
        start_of_day_position : Series
            Whether Long / Short or Flat at the start of the day.
        trade_action : Series
            Whether to Buy / Sell at the open.
        end_of_day_position : Series
            Whether Long / Short or Flat at the end of the day.

        """
        # Extract the trade signal from the OHLC Data
        eod_trade_signal = np.array(signal)

        # Create empty arrays to store data
        start_of_day_position = np.array([0]*len(prices))
        trade_action = np.array([0]*len(prices))
        end_of_day_position = np.array([0]*len(prices))

        # For each valid row in the data
        for row in range(start + 1, len(prices)):

            # The start of day position is equal to the close of day position
            # of the previous day
            start_of_day_position[row] = end_of_day_position[row-1]

            # The trade action is the previous days trade signal multiplied by
            # the position size
            trade_action[row] = eod_trade_signal[row-1]

            # The end of day position is the start of day position plus any
            # trade action
            end_of_day_position[row] = (start_of_day_position[row]
                                        + trade_action[row])

        pos_dict = {}
        pos_dict['start_of_day_position'] = start_of_day_position
        pos_dict['trade_action'] = trade_action
        pos_dict['end_of_day_position'] = end_of_day_position

        return pos_dict


    @staticmethod
    def position_values(
            prices, end_of_day_position, trade_price_dict):
        """
        Calculate position values

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_entry_price : Series
            The entry price for each trade.
        end_of_day_position : Series
            The number of units of position held at the end of day.
        trade_high_price : Series
            The high price for each trade.
        trade_low_price : Series
            The low price for each trade.
        trade_close_high_price : Series
            The highest closing price for each trade.
        trade_close_low_price : Series
            The lowest closing price for each trade.
        trade_number : Series
            Array of trade numbers.

        Returns
        -------
        initial_position_value : Series
            The MTM value of the position at trade entry.
        current_position_value : Series
            The MTM value of the position at the current date.
        max_trade_position_value : Series
            The maximum MTM value of the position up to the current date.
        max_trade_close_position_value : Series
            The maximum MTM closing value of the position up to the current
            date.
        min_trade_position_value : Series
            The minimum MTM value of the position up to the current date.
        min_trade_close_position_value : Series
            The minimum MTM closing value of the position up to the current
            date.

        """

        # Create a empty arrays to store the signals
        initial_position_value = np.array([0]*len(prices))
        current_position_value = np.array([0]*len(prices))
        max_trade_position_value = np.array([0]*len(prices))
        max_trade_close_position_value = np.array([0]*len(prices))
        min_trade_position_value = np.array([0]*len(prices))
        min_trade_close_position_value = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if prices['trade_number'][row] != 0:

                # If the end of day position is flat and the trade number is
                # the same as the previous day - i.e. the trade has just been
                # closed out
                if ((end_of_day_position[row] == 0) and (
                        prices['trade_number'][row] == prices[
                            'trade_number'][row-1])):

                    # Set the initial position value to the same as the
                    # previous day
                    initial_position_value[row] = initial_position_value[row-1]

                    # Set the current position value to the opening price
                    # multiplied by the end of day position of the previous day
                    current_position_value[row] = (
                        prices['Open'][row] * end_of_day_position[row-1])

                    # Set the maximum trade position value to the same as the
                    # previous day
                    max_trade_position_value[
                        row] = max_trade_position_value[row-1]

                    # Set the maximum trade closing position value to the same
                    # as the previous day
                    max_trade_close_position_value[
                        row] = max_trade_close_position_value[row-1]

                    # Set the minimum trade position value to the same as the
                    # previous day
                    min_trade_position_value[
                        row] = min_trade_position_value[row-1]

                    # Set the minimum trade closing position value to the same
                    # as the previous day
                    min_trade_close_position_value[
                        row] = min_trade_close_position_value[row-1]

                else:
                    # Set the initial position value to the trade entry price
                    # multiplied by the end of day position
                    initial_position_value[row] = (
                        trade_price_dict['trade_entry_price'][row]
                        * end_of_day_position[row])

                    # Set the current position value to the closing price
                    # multiplied by the end of day position
                    current_position_value[row] = (
                        prices['Close'][row] * end_of_day_position[row])

                    # Set the maximum trade position value to the high price of
                    # the trade multiplied by the end of day position
                    max_trade_position_value[row] = (
                        trade_price_dict['trade_high_price'][row]
                        * end_of_day_position[row])

                    # Set the maximum trade closing position value to the
                    # highest closing price of the trade multiplied by the end
                    # of day position
                    max_trade_close_position_value[row] = (
                        trade_price_dict['trade_close_high_price'][row]
                        * end_of_day_position[row])

                    # Set the minimum trade position value to the low price of
                    # the trade multiplied by the end of day position
                    min_trade_position_value[row] = (
                        trade_price_dict['trade_low_price'][row]
                        * end_of_day_position[row])

                    # Set the minimum trade closing position value to the
                    # lowest closing price of the trade multiplied by the end
                    # of day position
                    min_trade_close_position_value[row] = (
                        trade_price_dict['trade_close_low_price'][row]
                        * end_of_day_position[row])

        # Collect results in dictionary
        pos_val_dict = {}

        pos_val_dict['initial_position_value'] = initial_position_value
        pos_val_dict['current_position_value'] = current_position_value
        pos_val_dict['max_trade_position_value'] = max_trade_position_value
        pos_val_dict[
            'max_trade_close_position_value'] = max_trade_close_position_value
        pos_val_dict['min_trade_position_value'] = min_trade_position_value
        pos_val_dict[
            'min_trade_close_position_value'] = min_trade_close_position_value

        return pos_val_dict


    @classmethod
    def position_size(cls, prices, benchmark, params):
        """
        Calculate trade position size

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        benchmark : Series
            The closing prices of the benchmark series
        params : Dict
            equity : Float
                The account equity.
            start : Int
                Row from which trade signals can be generated.

        Returns
        -------
        position_size : Series
            The number of units of the chosen ticker to trade.
        benchmark_position_size : Series
            The number of units of the benchmark to trade.

        """
        # Initialize zero arrays to store position sizes
        prices['position_size'] = np.array(
            [0]*len(prices['Close']), dtype=int)
        prices['position_size_pp'] = np.array(
            [0]*len(prices['Close']), dtype=int)
        benchmark['benchmark_position_size'] = np.array(
            [0]*len(benchmark['Close']), dtype=int)

        # Get the date of the first trade
        trades = prices['raw_trade_number']
        params['first_trade_date'] = trades.loc[trades != 0].index[0]

        # Find the location of the start of the first trade
        params['first_trade_start'] = prices.index.get_loc(
            params['first_trade_date'])

        # Calculate position size based on initial account equity and first
        # trade closing prices
        if params['position_type'] == 'equity_constant':
            prices, params = cls._equity_constant_position_size(
                prices=prices, params=params)

        # Calculate position size based on initial account equity and closing
        # prices for each trade
        elif params['position_type'] == 'equity_variable':
            prices, params = cls._equity_variable_position_size(
                prices=prices, params=params)

        # Calculate position size based on proportion of ATR
        elif params['position_type'] == 'atr':
            prices, params = cls._atr_position_size(
                prices=prices, params=params)

        # Calculate position size as a fixed number of units
        else:
            prices, params = cls._fixed_position_size(
                prices=prices, params=params)

        # Calculate benchmark position size
        benchmark, params = cls._benchmark_position_size(
            benchmark=benchmark, params=params)

        # Set the initial position size to that of the first trade
        params['init_position_size'] = prices[
            'position_size'].loc[params['first_trade_date']]

        # Set the initial benchmark position size to that calculated as of the
        # same date
        params['init_benchmark_position_size'] = benchmark[
            'benchmark_position_size'].loc[params['first_trade_date']]

        return prices, benchmark, params


    @staticmethod
    def _benchmark_position_size(benchmark, params):

        # Set the position size to the number of shares that can be bought with
        # the initial equity

        # Set the number of units to use a percentage of starting equity at
        # the point when trade signals begin
        benchmark_units = math.floor(
            (params['equity']
             / benchmark['Close'].loc[params['first_trade_date']])
            * params['equity_inv_perc'])

        # Set the position size series to the number of units, starting from
        # the point that trade signals can be generated
        benchmark['benchmark_position_size'][
            benchmark.index.get_loc(
                params['first_trade_date']):] = benchmark_units

        return benchmark, params


    @staticmethod
    def _equity_constant_position_size(prices, params):

        # Extract the raw trade signal from the OHLC data
        trade_number = prices['raw_trade_number']

        # Set the position size to the number of shares that can be
        # bought with the initial equity

        # Set the number of units to use a percentage of starting equity at
        # the point when trade signals begin
        units = math.floor(
            (params['equity']
             / prices['Close'].loc[params['first_trade_date']])
            * params['equity_inv_perc']
            / params['contract_point_value'])

        # For each row since the first trade entry
        for row in range(params['first_trade_start'], len(prices['Close'])):

            # If the is a trade on
            if trade_number[row] != 0:

                # Set the position size series to the number of units
                prices['position_size'][row] = units

                # Set the position size for the perfect profit calc to the
                # same as the position size
                prices['position_size_pp'][row] = prices['position_size'][row]

            # If there is no trade on
            else:
                # Set the position size to zero
                prices['position_size'][row] = 0

                # Set the position size for the perfect profit calc to the
                # same as the previous day
                prices['position_size_pp'][row] = prices[
                    'position_size_pp'][row-1]

        return prices, params


    @staticmethod
    def _equity_variable_position_size(prices, params):

        # Extract the raw trade signal from the OHLC data
        trade_number = prices['raw_trade_number']

        # For each row since the first trade entry
        for row in range(params['first_trade_start'], len(prices['Close'])):

            # If the is a trade on
            if trade_number[row] != 0:

                # Get the index location of the trade entry date
                trade_first_row = prices.index.get_loc(
                    prices[trade_number==trade_number[row]].index[0])

                # If it is the trade entry date
                if row == trade_first_row:

                    # Set the number of units to use a percentage of starting
                    # equity at the point when trade signals begin
                    prices['position_size'][row] = math.floor(
                        (params['equity'] / prices['Close'][row])
                        * params['equity_inv_perc']
                        / params['contract_point_value'])

                # For every other day in the trade take the entry size
                else:
                    prices['position_size'][row] = prices[
                        'position_size'][row-1]

                # Set the position size for the perfect profit calc to the
                # same as the position size
                prices['position_size_pp'][row] = prices['position_size'][row]

            # If there is no trade on, set the position size to zero.
            else:
                prices['position_size'][row] = 0

                # Set the position size for the perfect profit calc to the
                # same as the previous day
                prices['position_size_pp'][row] = prices[
                    'position_size_pp'][row-1]

        return prices, params


    @staticmethod
    def _atr_position_size(prices, params):

        # Calculate ATR levels
        prices['position_ATR'] = Indicators.ATR(
            prices['High'], prices['Low'], prices['Close'],
            params['atr_pos_size'])

        # Extract the raw trade signal from the OHLC data
        trade_number = prices['raw_trade_number']

        # For each row since the first trade entry
        for row in range(params['first_trade_start'], len(prices['Close'])):

            # If the is a trade on
            if trade_number[row] != 0:

                # Get the index location of the trade entry date
                trade_first_row = prices.index.get_loc(
                    prices[trade_number==trade_number[row]].index[0])

                # If it is the trade entry date
                if row == trade_first_row:

                    # If we can calculate the ATR
                    if row > params['atr_pos_size']:

                        # Size the position for each trade based on a fraction
                        # of the ATR
                        prices['position_size'][row] = math.floor(
                            (params['equity'] * (params['position_risk_bps']
                                                 / 10000))
                            / (prices['position_ATR'][row]
                               * params['contract_point_value']))

                    # Otherwise
                    else:
                        # Set the position size to 1
                        prices['position_size'][row] = 1

                # For every other day in the trade take the entry size
                else:
                    prices['position_size'][row] = prices[
                        'position_size'][row-1]

                # Set the position size for the perfect profit calc to the
                # same as the position size
                prices['position_size_pp'][row] = prices['position_size'][row]

            # If there is no trade on
            else:
                # Set the position size to zero.
                prices['position_size'][row] = 0

                # Set the position size for the perfect profit calc to the
                # same as the previous day
                prices['position_size_pp'][row] = prices[
                    'position_size_pp'][row-1]

        return prices, params


    @staticmethod
    def _fixed_position_size(prices, params):

        # Extract the raw trade signal from the OHLC data
        trade_number = prices['raw_trade_number']

        units = params['fixed_pos_size']

        # For each row since the first trade entry
        for row in range(params['first_trade_start'], len(prices['Close'])):

            # If the is a trade on
            if trade_number[row] != 0:
                prices['position_size'][row] = units

            # If there is no trade on
            else:
                prices['position_size'][row] = 0

            # Set position size for perfect profit calculation
            prices['position_size_pp'][row] = units

        return prices, params



"""
Entry and Exit signals

"""

from tradingsystems.dollar_exits import DollarExit
from tradingsystems.indicator_entries import IndicatorEntry
from tradingsystems.indicator_exits import IndicatorExit
from tradingsystems.ma_entries import MovingAverageEntry

class Signals():
    """
    Calculate entry signals

    """

    @classmethod
    def entry_signal(cls, tables, params):
        """
        Calculate trade entry signals

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        entry_type : Str, optional
            The entry strategy. The default is '2ma'.
        ma1 : Int, optional
            The first moving average period.
        ma2 : Int, optional
            The second moving average period.
        ma3 : Int, optional
            The third moving average period.
        ma4 : Int, optional
            The fourth moving average period.
        simple_ma : Bool
            Whether to calculate a simple or exponential moving average. The
            default is True.
        entry_period : Int
            The number of days to use in the entry strategy. The default is 14.
        entry_oversold : Int
            The oversold level to use in the entry strategy.
        entry_overbought : Int
            The overbought level to use in the entry strategy.
        entry_threshold : Float
            The entry threshold used for momentum / volatility strategies.
            The default is 0 for momentum and 1.5 for volatility.
        entry_acceleration_factor : Float
            The acceleration factor used in the Parabolic SAR entry signal.
            The default is 0.02.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        start : Int
            The first valid row to start calculating trade information from.
        signal : Series
            The series of Buy / Sell signals.

        """

        # Double Moving Average Crossover
        if params['entry_type'] == '2ma':
            tables['prices'], start, \
                signal = MovingAverageEntry.entry_double_ma_crossover(
                    prices=tables['prices'],
                    params=params)

        # Triple Moving Average Crossover
        elif params['entry_type'] == '3ma':
            tables['prices'], start, \
                signal = MovingAverageEntry.entry_triple_ma_crossover(
                    prices=tables['prices'],
                    params=params)

        # Quad Moving Average Crossover
        elif params['entry_type'] == '4ma':
            tables['prices'], start, \
                signal = MovingAverageEntry.entry_quad_ma_crossover(
                    prices=tables['prices'],
                    params=params)

        # Parabolic SAR
        elif params['entry_type'] == 'sar':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_parabolic_sar(
                    prices=tables['prices'],
                    acceleration_factor=params['entry_acceleration_factor'])

        # Channel Breakout
        elif params['entry_type'] == 'channel_breakout':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_channel_breakout(
                    prices=tables['prices'],
                    time_period=params['entry_period'])

        # Stochastic Crossover
        elif params['entry_type'] == 'stoch_cross':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_stochastic_crossover(
                    tables['prices'],
                    time_period=params['entry_period'],
                    oversold=params['entry_oversold'],
                    overbought=params['entry_overbought'])

        # Stochastic Over Under
        elif params['entry_type'] == 'stoch_over_under':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_stochastic_over_under(
                    tables['prices'],
                    time_period=params['entry_period'],
                    oversold=params['entry_oversold'],
                    overbought=params['entry_overbought'])

        # Stochastic Pop
        elif params['entry_type'] == 'stoch_pop':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_stochastic_pop(
                    tables['prices'],
                    time_period=params['entry_period'],
                    oversold=params['entry_oversold'],
                    overbought=params['entry_overbought'])

        # Relative Strength Index
        elif params['entry_type'] == 'rsi':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_rsi(
                    tables['prices'],
                    time_period=params['entry_period'],
                    oversold=params['entry_oversold'],
                    overbought=params['entry_overbought'])

        # Commodity Channel Index
        elif params['entry_type'] == 'cci':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_commodity_channel_index(
                    tables['prices'],
                    time_period=params['entry_period'],
                    threshold=params['entry_threshold'])

        # Momentum
        elif params['entry_type'] == 'momentum':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_momentum(
                    tables['prices'],
                    time_period=params['entry_period'],
                    threshold=params['entry_threshold'])

        # Volatility
        elif params['entry_type'] == 'volatility':
            tables['prices'], start, \
                signal = IndicatorEntry.entry_volatility(
                    tables['prices'],
                    time_period=params['entry_period'],
                    threshold=params['entry_threshold'])

        return tables['prices'], start, signal


    @classmethod
    def exit_and_stop_signals(cls, prices, params):
        """
        Calculate trade exit and stop signals.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        sip_price : Bool
            Whether to set the SIP of the Parabolic SAR exit to n-day
            high / low or to the high of the previous trade. The default is
            False.
        exit_type : Str, optional
            The exit strategy. The default is 'trailing_stop'.
        exit_period : Int, optional
            The number of days to use in the exit strategy. The default is 5.
        stop_type : Str, optional
            The stop strategy. The default is 'initial_dollar'.
        stop_period : Int, optional
            The number of days to use in the stop strategy. The default is 5.
        exit_threshold : Float
            The exit threshold used for the volatility strategy.
            The default is 1.
        exit_oversold : Int, optional
            The oversold level to use in the exit strategy.
        exit_overbought : Int, optional
            The overbought level to use in the exit strategy.
        exit_acceleration_factor : Float
            The acceleration factor used in the Parabolic SAR exit signal.
            The default is 0.02.
        exit_trailing_close : Series, optional
            The exit levels for each trade based on the trailing close
        exit_profit_target : Series, optional
            The exit levels for each trade based on a profit target
        stop_initial_dollar_loss : Series, optional
            The stop levels for each trade based on a dollar loss from the
            entry level.
        stop_profit_target : Series, optional
            The stop levels for each trade based on a profit target
        stop_trailing_close : Series, optional
            The stop levels for each trade based on the trailing close
        stop_trailing_high_low : Series, optional
            The stop levels for each trade based on the trailing high / low.

        Returns
        -------
        prices : DataFrame
            The OHLC data

        """
        # Generate the exit signals
        prices, prices['exit_signal'] = cls._exit_signal(
            prices=prices, params=params)

        # Generate the stop signals
        prices, prices['stop_signal'] = cls._stop_signal(
            prices=prices, params=params)

        return prices


    @classmethod
    def _exit_signal(cls, prices, params):
        """
        Calculate trade exit signals.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        exit_type : Str, optional
            The exit strategy. The default is 'trailing_stop'.
        exit_period : Int, optional
            The number of days to use in the exit strategy. The default is 5.
        exit_threshold : Float
            The exit threshold used for the volatility strategy.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        exit_oversold : Int, optional
            The oversold level to use in the exit strategy.
        exit_overbought : Int, optional
            The overbought level to use in the exit strategy.
        exit_acceleration_factor : Float
            The acceleration factor used in the Parabolic SAR exit signal.
            The default is 0.02.
        sip_price : Bool
            Whether to set the SIP of the Parabolic SAR exit to n-day
            high / low or to the high of the previous trade. The default is
            False.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        exit : Series
            The exit signals.


        prices : DataFrame
            The OHLC data

        """

        # Parabolic SAR Exit
        if params['exit_type'] == 'sar':
            prices, exit_ = IndicatorExit.exit_parabolic_sar(
                prices=prices,
                time_period=params['exit_period'],
                acceleration_factor=params['exit_acceleration_factor'],
                sip_price=params['sip_price'])

        # Support / Resistance Exit
        elif params['exit_type'] == 'sup_res':
            prices, exit_ = IndicatorExit.exit_support_resistance(
                prices=prices,
                time_period=params['exit_period'])

        # Trailing RSI Exit
        elif params['exit_type'] == 'rsi_trail':
            prices, exit_ = IndicatorExit.exit_rsi_trail(
                prices=prices,
                time_period=params['exit_period'],
                oversold=params['exit_oversold'],
                overbought=params['exit_overbought'])

        # Key Reversal Day Exit
        elif params['exit_type'] == 'key_reversal':
            prices, exit_ = IndicatorExit.exit_key_reversal(
                prices=prices,
                time_period=params['exit_period'])

        # Volatility Breakout Exit
        elif params['exit_type'] == 'volatility':
            prices, exit_ = IndicatorExit.exit_volatility(
                prices=prices,
                time_period=params['exit_period'],
                threshold=params['exit_threshold'])

        # Stochastic Crossover Exit
        elif params['exit_type'] == 'stoch_cross':
            prices, exit_ = IndicatorExit.exit_stochastic_crossover(
                prices=prices,
                time_period=params['exit_period'])

        # N-day Range Exit
        elif params['exit_type'] == 'nday_range':
            prices, exit_ = IndicatorExit.exit_nday_range(
                prices=prices,
                time_period=params['exit_period'])

        # Random Exit
        elif params['exit_type'] == 'random':
            prices, exit_ = IndicatorExit.exit_random(
                prices=prices)

        # Trailing Stop Exit
        elif params['exit_type'] == 'trailing_stop':
            prices, exit_ = DollarExit.exit_dollar(
                exit_level='trail_close',
                prices=prices,
                trigger_value=prices['exit_trailing_close'])

        # Profit Target Exit
        elif params['exit_type'] == 'profit_target':
            prices, exit_ = DollarExit.exit_dollar(
                exit_level='profit_target',
                prices=prices,
                trigger_value=prices['exit_profit_target'])

        return prices, exit_


    @classmethod
    def _stop_signal(cls, prices, params):
        """
        Calculate trade stop signals

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        stop_type : Str
            The type of stop to use.
        stop_period : Int
            The length of time for the indicator calculation.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        stop_initial_dollar_loss : Series, optional
            The stop levels for each trade based on a dollar loss from the
            entry level.
        stop_profit_target : Series, optional
            The stop levels for each trade based on a profit target
        stop_trailing_close : Series, optional
            The stop levels for each trade based on the trailing close
        stop_trailing_high_low : Series, optional
            The stop levels for each trade based on the trailing high / low.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        stop : Series
            The stop signals.

        """

        # Support / Resistance Stop
        if params['stop_type'] == 'sup_res':
            prices, stop = IndicatorExit.exit_support_resistance(
                prices=prices,
                time_period=params['stop_period'])

        # Immediate Profit Stop
        elif params['stop_type'] == 'immediate_profit':
            prices, stop = IndicatorExit.exit_immediate_profit(
                prices=prices,
                time_period=params['stop_period'])

        # Initial Dollar Loss Stop
        elif params['stop_type'] == 'initial_dollar':
            prices, stop = DollarExit.exit_dollar(
                exit_level='initial',
                prices=prices,
                trigger_value=prices['stop_initial_dollar_loss'])

        # Breakeven Stop
        elif params['stop_type'] == 'breakeven':
            prices, stop = DollarExit.exit_dollar(
                exit_level='breakeven',
                prices=prices,
                trigger_value=prices['stop_profit_target'])

        # Trailing Stop (Closing Price)
        elif params['stop_type'] == 'trail_close':
            prices, stop = DollarExit.exit_dollar(
                exit_level='trail_close',
                prices=prices,
                trigger_value=prices['stop_trailing_close'])

        # Trailing Stop (High / Low Price)
        elif params['stop_type'] == 'trail_high_low':
            prices, stop = DollarExit.exit_dollar(
                exit_level='trail_high_low',
                prices=prices,
                trigger_value=prices['stop_trailing_high_low'])

        return prices, stop



"""
Exit signals based on a dollar value.

"""
import numpy as np

class DollarExit():
    """
    Calculate dollar value based exit signals.

    """

    @classmethod
    def exit_dollar(
            cls, prices, trigger_value, exit_level):
        """
        Calculate exit based on a dollar amount.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trigger_value : Series
            The series to trigger exit.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        trade_high_price : Series
            The high price of the trade.
        trade_low_price : Series
            The low price of the trade.
        exit_level : Str
            The type of exit strategy.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        exit : Series
            Exit signals.

        """

        # Calculate exit signal based on a profit target
        if exit_level == 'profit_target':
            prices, exit_ = cls._exit_profit_target(
                prices=prices,
                trigger_value=trigger_value)

        # Calculate exit signal based on a loss from entry price
        elif exit_level == 'initial':
            prices, exit_ = cls._exit_initial_dollar_loss(
                prices=prices,
                trigger_value=trigger_value)

        # Calculate exit signal based on a breakeven level
        elif exit_level == 'breakeven':
            prices, exit_ = cls._exit_breakeven(
                prices=prices,
                trigger_value=trigger_value)

        # Calculate exit signal based on a trailing stop referencing the close
        elif exit_level == 'trail_close':
            prices, exit_ = cls._exit_trailing(
                prices=prices,
                trigger_value=trigger_value)

        # Calculate exit signal based on a trailing stop referencing the
        # high/low
        elif exit_level == 'trail_high_low':
            prices, exit_ = cls._exit_trailing(
                prices=prices,
                trigger_value=trigger_value)

        return prices, exit_


    @staticmethod
    def _exit_profit_target(prices, trigger_value):
        """
        Calculate exit based on a profit target.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trigger_value : Series
            The series to trigger exit.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        profit_target_exit : Series
            The exit signals.

        """
        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Create an empty array to store the signals
        profit_target_exit = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the close is greater than the trigger value
                    if prices['Close'][row] > trigger_value[row]:

                        # Set the exit signal to -1
                        profit_target_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the close is less than the trigger value
                    if prices['Close'][row] < trigger_value[row]:

                        # Set the exit signal to 1
                        profit_target_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    profit_target_exit[row] = 0

        return prices, profit_target_exit


    @staticmethod
    def _exit_initial_dollar_loss(prices, trigger_value):
        """
        Calculate exit based on a given loss from the entry point.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trigger_value : Series
            The series to trigger exit.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        initial_dollar_loss_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Create an empty array to store the signals
        initial_dollar_loss_exit = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the close is less than the trigger value
                    if prices['Close'][row] < trigger_value[row]:

                        # Set the exit signal to -1
                        initial_dollar_loss_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the close is greater than the trigger value
                    if prices['Close'][row] > trigger_value[row]:

                        # Set the exit signal to 1
                        initial_dollar_loss_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    initial_dollar_loss_exit[row] = 0

        return prices, initial_dollar_loss_exit


    @staticmethod
    def _exit_breakeven(prices, trigger_value):
        """
        Calculate exit based on passing a breakeven threshold.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trigger_value : Series
            The series to trigger exit.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        trade_high_price : Series
            The high price of the trade.
        trade_low_price : Series
            The low price of the trade.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        breakeven_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']
        trade_high_price = prices['raw_trade_high_price']
        trade_low_price = prices['raw_trade_low_price']

        # Create an empty array to store the signals
        breakeven_exit = np.array([0.0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the high price of the trade is greater than the
                    # trigger value
                    if trade_high_price[row] > trigger_value[row]:

                        # If the close is less than the trigger value
                        if prices['Close'][row] < trigger_value[row]:

                            # Set the exit signal to -1
                            breakeven_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the low price of the trade is less than the
                    # trigger value
                    if trade_low_price[row] < trigger_value[row]:

                        # If the close is greater than the trigger value
                        if prices['Close'][row] > trigger_value[row]:

                            # Set the exit signal to 1
                            breakeven_exit[row] = 1
                else:
                    # Set the exit signal to 0
                    breakeven_exit[row] = 0

        return prices, breakeven_exit


    @staticmethod
    def _exit_trailing(prices, trigger_value):
        """
        Calculate exit based on a trailing stop.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trigger_value : Series
            The series to trigger exit.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        trailing_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Create an empty array to store the signals
        trailing_exit = np.array([0.0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the close is less than the trigger value
                    if prices['Close'][row] < trigger_value[row]:

                        # Set the exit signal to -1
                        trailing_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the close is greater than the trigger value
                    if prices['Close'][row] > trigger_value[row]:

                        # Set the exit signal to 1
                        trailing_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    trailing_exit[row] = 0

        return prices, trailing_exit



"""
Calculate performance for various technical trading systems and display the
results in table and graph form.

"""

# Imports
import copy
import pandas as pd
from tradingsystems.graphs import PerformanceGraph as perfgraph
from tradingsystems.marketdata import Markets
from tradingsystems.positions import Positions
from tradingsystems.pnl import Profit
from tradingsystems.reports import PerfReport
from tradingsystems.signals import Signals
from tradingsystems.targets import TradeTargets
from tradingsystems.trades import Trades
from tradingsystems.systems_params import system_params_dict
from tradingsystems.utils import Labels, Dates, Reformat


class TestStrategy():
    """
    Run a backtest over the chosen strategy

    Parameters
    ----------
    ticker : Str, optional
        Underlying to test. The default '$SPX'.
    ccy_1 : Str, optional
        Primary currency of pair to return. The default 'GBP'.
    ccy_2 : Str, optional
        Secondary currency of pair to return. The default 'USD'.
    asset_type : Str
        The alphavantage asset class type. The default is 'fx'.
    start_date : Str, optional
        Date to begin backtest. Format is YYYY-MM-DD.
    end_date : Str, optional
        Date to end backtest. Format is YYYY-MM-DD.
    lookback : Int, optional
        Number of business days to use for the backtest. The default is 750
        business days (circa 3 years).
    ma1 : Int, optional
        The first moving average period.
    ma2 : Int, optional
        The second moving average period.
    ma3 : Int, optional
        The third moving average period.
    ma4 : Int, optional
        The fourth moving average period.
    simple_ma : Bool, optional
        Whether to calculate a simple or exponential moving average. The
        default is True.
    position_size : Int, optional
        The number of units to trade. The default is based on equity.
    pos_size_fixed : Bool
        Whether to used a fixed position size for all trades. The default
        is True.
    ticker_source : Str, optional
        The data source to use for the ticker data, either 'norgate',
        'alpha' or 'yahoo'. The default is 'norgate'.
    bench_ticker : Str, optional
        Underlying to use as benchmark. The default '$SPX'.
    bench_source : Str, optional
        The data source to use for the benchmark data, either 'norgate',
        'alpha' or 'yahoo'. The default is 'norgate'.
    slippage : Float, optional
        The amount of slippage to apply to traded prices in basis points.
        The default is 5 bps per unit.
    commission : Float, optional
        The amount of commission charge to apply to each trade. The
        default is $0.00.
    riskfree : Float, optional
        The riskfree interest rate. The default is 25bps.
    entry_type : Str, optional
        The entry strategy. The default is '2ma'.
    exit_type : Str, optional
        The exit strategy. The default is 'trailing_stop'.
    stop_type : Str, optional
        The stop strategy. The default is 'initial_dollar'.
    entry_period : Int, optional
        The number of days to use in the entry strategy. The default is 14.
    exit_period : Int, optional
        The number of days to use in the exit strategy. The default is 5.
    stop_period : Int, optional
        The number of days to use in the stop strategy. The default is 5.
    entry_oversold : Int, optional
        The oversold level to use in the entry strategy.
    entry_overbought : Int, optional
        The overbought level to use in the entry strategy.
    exit_oversold : Int, optional
        The oversold level to use in the exit strategy.
    exit_overbought : Int, optional
        The overbought level to use in the exit strategy.
    entry_threshold : Float
        The entry threshold used for momentum / volatility strategies.
        The default is 0 for momentum and 1.5 for volatility.
    exit_threshold : Float
        The exit threshold used for the volatility strategy.
        The default is 1.
    entry_acceleration_factor : Float
        The acceleration factor used in the Parabolic SAR entry signal.
        The default is 0.02.
    exit_acceleration_factor : Float
        The acceleration factor used in the Parabolic SAR exit signal.
        The default is 0.02.
    sip_price : Bool
        Whether to set the SIP of the Parabolic SAR exit to n-day
        high / low or to the high of the previous trade. The default is
        False.
    equity : Float
        The initial account equity level. The default is $100,000.00.
    exit_amount : Float
        The dollar exit amount. The default is $1000.00.
    stop_amount : Float
        The dollar stop amount. The default is $500.00.
    api_key : Str
        AlphaVantage API key. If not provided will look for
        'ALPHAVANTAGE_API_KEY' in the environment variables.

    Returns
    -------
    Results
        Prints out performance data for the strategy.

    """

    def __init__(self, **kwargs):

        # Import dictionary of default parameters
        self.default_dict = copy.deepcopy(system_params_dict)

        # Longnames for Norgate Tickers
        self.norgate_name_dict = Markets.norgate_name_dict()

        # Store initial inputs
        inputs = {}
        for key, value in kwargs.items():
            inputs[key] = value

        # Initialise system parameters
        params = self._init_params(inputs)

        # Set the start and end dates if not provided
        params['start_date'], params['end_date'] = Dates.date_set(
            start_date=params['start_date'], end_date=params['end_date'],
            lookback=params['lookback'])

        # Create DataFrame of OHLC prices from NorgateData or Yahoo Finance
        tables = {}

        tables['prices'], params = Markets.create_base_data(
            ticker=params['ticker'], source=params['ticker_source'],
            params=params, bench_flag=False)

        # Extract benchmark data for Beta calculation
        tables['benchmark'], params = Markets.create_base_data(
            ticker=params['bench_ticker'], source=params['bench_source'],
            params=params, bench_flag=True)

        # Set the strategy labels
        labels  = {}
        labels['entry_label'], labels['exit_label'], \
            labels['stop_label'] = Labels.strategy_labels(
                params=params, default_dict=self.default_dict)

        # Generate initial trade data
        tables, params, raw_trade_price_dict = self._raw_entry_signals(
            tables=tables, params=params)

        # Create exit and stop targets
        tables['prices'] = TradeTargets.exit_and_stop_targets(
            prices=tables['prices'], params=params,
            trade_price_dict=raw_trade_price_dict)

        # Create exit and stop signals
        tables['prices'] = Signals.exit_and_stop_signals(
            prices=tables['prices'], params=params)

        # Concatenate the Entry, Exit and Stop signals in a single DataFrame
        trade_signals = pd.concat(
            [tables['prices']['raw_trade_signal'],
             tables['prices']['exit_signal'],
             tables['prices']['stop_signal']], axis=1)

        # Generate single combined trade signal
        tables['prices']['combined_signal'] = Trades.signal_combine(
            prices=tables['prices'], start=params['start'],
            end_of_day_position=tables['prices']['raw_end_of_day_position'],
            trade_signals=trade_signals)

        # Create trade and position data
        pos_dict = Positions.calc_positions(
            prices=tables['prices'],
            signal=tables['prices']['combined_signal'],
            start=params['start'])

        # Scale the position info by the position size
        pos_dict = Reformat.position_scale(
            pos_dict=pos_dict, position_size=tables['prices']['position_size'])

        # Map the raw positions to the OHLC data
        tables['prices'] = Reformat.map_to_prices(
            prices=tables['prices'],
            input_dict=pos_dict,
            title_modifier='')

        tables['prices']['trade_number'] = Trades.trade_numbers(
            prices=tables['prices'],
            end_of_day_position=tables['prices']['end_of_day_position'],
            start=params['start'])

        # Calculate the trades and pnl for the strategy
        tables['prices'] = Profit.profit_data(
            prices=tables['prices'],
            params=params)

        # Create monthly summary data
        tables['monthly_data'] = Profit.create_monthly_data(
            prices=tables['prices'], equity=params['equity'])

        # Create dictionary of performance data
        tables['perf_dict'] = PerfReport.performance_data(
            tables=tables, params=params, labels=labels,
            norgate_name_dict=self.norgate_name_dict,)

        # Print out results
        self.performance_report(tables=tables)

        self.params = params
        self.tables = tables
        self.labels = labels
        self.inputs = inputs


    @staticmethod
    def _init_params(inputs):
        """
        Initialise parameter dictionary

        Parameters
        ----------
        inputs : Dict
            Dictionary of parameters supplied to the function.

        Returns
        -------
        params : Dict
            Dictionary of parameters.

        """
        # Copy the default parameters
        params = copy.deepcopy(system_params_dict['df_params'])

        # Extract the entry, exit and stop signal dictionaries
        entry_signal_dict = system_params_dict['df_entry_signal_dict']
        exit_signal_dict = system_params_dict['df_exit_signal_dict']
        stop_signal_dict = system_params_dict['df_stop_signal_dict']

        # For all the supplied arguments
        for key, value in inputs.items():

            # Replace the default parameter with that provided
            params[key] = value

        # Set the start and end dates to None if not supplied
        if 'start_date' not in inputs.keys():
            params['start_date'] = None

        if 'end_date' not in inputs.keys():
            params['end_date'] = None

        # Create a list of the entry, exit and stop types
        types = [
            params['entry_type'], params['exit_type'], params['stop_type']]

        # For each parameter in params
        for param in params.keys():

            # If the parameter has not been supplied as an input and it is not
            # the entry exit or stop type
            if (param not in inputs.keys()
                and param not in types):

                # If the parameter takes a specific value for the particular
                # entry type then replace the default with this value
                if param in entry_signal_dict[types[0]].keys():
                    params[param] = entry_signal_dict[types[0]][str(param)]

                # If the parameter takes a specific value for the particular
                # exit type then replace the default with this value
                if param in exit_signal_dict[types[1]].keys():
                    params[param] = exit_signal_dict[types[1]][str(param)]

                # If the parameter takes a specific value for the particular
                # stop type then replace the default with this value
                if param in stop_signal_dict[types[2]].keys():
                    params[param] = stop_signal_dict[types[2]][str(param)]

        return params


    @classmethod
    def _raw_entry_signals(cls, tables, params):
        """
        Generate the initial raw entry signals, positions and trades

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        entry_type : Str, optional
            The entry strategy. The default is '2ma'.
        ma1 : Int, optional
            The first moving average period.
        ma2 : Int, optional
            The second moving average period.
        ma3 : Int, optional
            The third moving average period.
        ma4 : Int, optional
            The fourth moving average period.
        entry_period : Int
            The number of days to use in the entry strategy. The default is 14.
        entry_oversold : Int
            The oversold level to use in the entry strategy.
        entry_overbought : Int
            The overbought level to use in the entry strategy.
        entry_threshold : Float
            The entry threshold used for momentum / volatility strategies.
            The default is 0 for momentum and 1.5 for volatility.
        simple_ma : Bool
            Whether to calculate a simple or exponential moving average. The
            default is True.
        benchmark : Series
            The series of closing prices of the benchmark underlying.
        entry_acceleration_factor : Float
            The acceleration factor used in the Parabolic SAR entry signal.
            The default is 0.02.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        start : Int
            The first valid row to start calculating trade information from.

        """
        # Generate entry signals
        tables['prices'], params['start'], \
            tables['prices']['raw_trade_signal'] = Signals.entry_signal(
                tables=tables, params=params)

        # Calculate initial position info
        raw_pos_dict = Positions.calc_positions(
            prices=tables['prices'],
            signal=tables['prices']['raw_trade_signal'],
            start=params['start'])

        # Generate trade numbers
        tables['prices']['raw_trade_number'] = Trades.trade_numbers(
            prices=tables['prices'],
            end_of_day_position=raw_pos_dict['end_of_day_position'],
            start=params['start'])

        # Set the position size
        tables['prices'], tables['benchmark'], \
            params = Positions.position_size(
                prices=tables['prices'], benchmark=tables['benchmark'],
                params=params)

        # Set the position size label
        params = Labels.position_size_label(params)

        # Scale the position info by the position size
        raw_pos_dict = Reformat.position_scale(
            pos_dict=raw_pos_dict,
            position_size=tables['prices']['position_size'])

        # Generate initial trade prices
        raw_trade_price_dict = Trades.trade_prices(
            prices=tables['prices'],
            trade_number=tables['prices']['raw_trade_number'])

        # Map the raw positions to the OHLC data
        tables['prices'] = Reformat.map_to_prices(
            prices=tables['prices'],
            input_dict=raw_pos_dict,
            title_modifier='raw_')

        # Map the raw trade prices to the OHLC data
        tables['prices'] = Reformat.map_to_prices(
            prices=tables['prices'],
            input_dict=raw_trade_price_dict,
            title_modifier='raw_')

        return tables, params, raw_trade_price_dict


    @staticmethod
    def performance_report(tables):
        """
        Display the performance report

        Parameters
        ----------
        tables : Dict
            Dictionary containing performance dict.

        Returns
        -------
        Prints the performance report to the console.

        """
        # Print out results
        PerfReport.report_table(input_dict=tables['perf_dict'])


    def performance_graph(self, signals=None):
        """
        Graph the performance of the strategy

        Parameters
        ----------
        signals : Bool
            Whether to plot the Buy/Sell signals on the price chart.

        Returns
        -------
        Matplotlib Chart
            Displays the chart.

        """

        # If the signals flag is not provided, take the default value
        if signals is None:
            signals = self.default_dict['df_signals']

        # Dictionary to store entry signal data
        es_dict = {}

        # Entry labels
        es_dict['entry_signal_labels'] = self.default_dict[
            'df_entry_signal_labels']

        # Entry signal indicator column names
        es_dict['entry_signal_indicators'] = self.default_dict[
            'df_entry_signal_indicators']

        params = self.params
        #labels = self.labels
        tables = self.tables

        # For these entry strategies just plot the price chart and equity curve
        if params['entry_type'] in [
                '2ma', '3ma', '4ma', 'sar', 'channel_breakout']:
            graph = perfgraph.two_panel_graph(
                signals=signals, tables=tables, params=params,
                es_dict=es_dict)

        # Otherwise also plot the relevant indicator
        else:
            graph = perfgraph.three_panel_graph(
                signals=signals, tables=tables, params=params,
                es_dict=es_dict)

        return graph
